"""Helpers for the asset queries."""

import warnings
from concurrent.futures import ThreadPoolExecutor
from itertools import repeat
from mimetypes import guess_extension
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import requests
from tenacity import retry
from tenacity.stop import stop_after_attempt
from tenacity.wait import wait_random

from kili.core.graphql import QueryOptions
from kili.core.graphql.operations.data_connection.queries import (
    DataConnectionsQuery,
    DataConnectionsWhere,
)
from kili.core.graphql.operations.project.queries import ProjectQuery, ProjectWhere
from kili.exceptions import NotFound

from .exceptions import DownloadNotAllowedError, MissingPropertyError


def get_download_assets_function(
    kili,
    download_media: bool,
    fields: List[str],
    project_id: str,
    local_media_dir: Optional[str],
) -> Tuple[Optional[Callable], List[str]]:
    """Get the function to be called after each batch of asset query.

    The function is either None or MediaDownloader.download_assets().

    Also returns the fields to be queried, which may be modified
    if the jsonContent field is necessary.
    """
    if not download_media:
        return None, fields

    projects = list(
        ProjectQuery(kili.graphql_client)(
            ProjectWhere(project_id=project_id), ["inputType"], QueryOptions(disable_tqdm=True)
        )
    )
    if len(projects) == 0:
        raise NotFound(
            f"project ID: {project_id}. Maybe your KILI_API_KEY does not belong to a member of the"
            " project."
        )

    # We need to query the data connections to know if the assets are hosted in a cloud storage
    # If so, we remove the fields "content" and "jsonContent" from the query
    data_connections_gen = DataConnectionsQuery(kili.graphql_client)(
        where=DataConnectionsWhere(project_id=project_id),
        fields=["id"],
        options=QueryOptions(disable_tqdm=True, first=1, skip=0),
    )
    if len(list(data_connections_gen)) > 0:
        raise DownloadNotAllowedError(
            "The download of assets from a project connected to a cloud storage is not allowed."
            " Asset download is disabled."
        )

    input_type = projects[0]["inputType"]
    jsoncontent_field_added = False
    if input_type in ("TEXT", "VIDEO") and "jsonContent" not in fields:
        fields = fields + ["jsonContent"]
        jsoncontent_field_added = True

    return (
        MediaDownloader(
            local_media_dir,
            project_id,
            jsoncontent_field_added,
            input_type,
        ).download_assets,
        fields,
    )


class MediaDownloader:
    """Media downloader for kili.assets()."""

    def __init__(
        self,
        local_media_dir: Optional[Union[Path, str]],
        project_id: str,
        jsoncontent_field_added: bool,
        project_input_type: str,
    ) -> None:
        self.local_media_dir = local_media_dir
        self.project_id = project_id
        self.jsoncontent_field_added = jsoncontent_field_added
        self.project_input_type = project_input_type

        self.local_dir_path = (
            Path(self.local_media_dir)
            if self.local_media_dir is not None
            else Path.home() / ".cache" / "kili" / "projects" / self.project_id / "assets"
        )

    def download_assets(self, assets: List[Dict]) -> List[Dict]:
        """Download assets media in local."""
        if len(assets) == 0:
            return assets

        assert_required_fields_existence(assets)

        self.local_dir_path.mkdir(parents=True, exist_ok=True)

        with ThreadPoolExecutor() as threads:
            assets_gen = threads.map(self.download_single_asset, assets)

        assets = list(assets_gen)

        if self.jsoncontent_field_added:
            jsoncontent_not_empty = any(bool(asset["jsonContent"]) for asset in assets)
            if jsoncontent_not_empty:
                warnings.warn(
                    "Non empty jsonContent found in assets. Field was automatically added.",
                    stacklevel=2,
                )
            else:
                for asset in assets:
                    del asset["jsonContent"]

        return assets

    def download_single_asset(self, asset: Dict) -> Dict[str, Any]:
        """Download single asset on disk and modify asset attributes."""
        if "ocrMetadata" in asset and str(asset["ocrMetadata"]).startswith("http"):
            response = requests.get(asset["ocrMetadata"], timeout=20).json()
            asset["ocrMetadata"] = response

        if "jsonContent" in asset and str(asset["jsonContent"]).startswith("http"):
            # richtext
            if self.project_input_type == "TEXT":
                asset["jsonContent"] = download_file(
                    asset["jsonContent"], asset["externalId"], self.local_dir_path
                )

            # video frames
            elif self.project_input_type == "VIDEO":
                response = requests.get(asset["jsonContent"], timeout=20).json()
                urls = tuple(response.values())
                nbr_char_zfill = len(str(len(urls)))
                img_names = (
                    f'{asset["externalId"]}_{f"{i+1}".zfill(nbr_char_zfill)}'
                    for i, _ in enumerate(urls)
                )
                with ThreadPoolExecutor() as threads:
                    paths_gen = threads.map(
                        download_file,
                        urls,
                        img_names,
                        repeat(self.local_dir_path),
                    )
                    asset["jsonContent"] = list(paths_gen)
                return asset  # we skip video "content" download

            # big images
            elif self.project_input_type == "IMAGE":
                # the "jsonContent" contains some information but not the image
                response = requests.get(asset["jsonContent"], timeout=20).json()
                asset["jsonContent"] = response

            else:
                raise NotImplementedError(
                    f"jsonContent download for type {self.project_input_type} not implemented yet."
                )

        if str(asset["content"]).startswith("http"):
            asset["content"] = download_file(
                asset["content"], asset["externalId"], self.local_dir_path
            )
            return asset

        return asset


def get_file_extension_from_headers(url) -> Optional[str]:
    """Guess the extension of a file with the url response headers."""
    with requests.head(url, timeout=20) as header_response:
        if header_response.status_code == 200:
            headers = header_response.headers
        else:
            with requests.get(url, timeout=20) as response:
                response.raise_for_status()
                headers = response.headers
        if "content-type" in headers:
            content_type = headers["content-type"]
            return guess_extension(content_type)
    return None


def get_download_path(url: str, external_id: str, local_dir_path: Path) -> Path:
    """Build the path to download a file the file in local."""
    extension = get_file_extension_from_headers(url)
    filename = external_id
    if extension is not None and not filename.endswith(extension):
        filename = filename + extension
    local_dir_path = local_dir_path / filename
    return local_dir_path.resolve()


@retry(stop=stop_after_attempt(2), wait=wait_random(min=1, max=2), reraise=True)
def download_file(url: str, external_id: str, local_dir_path: Path) -> str:
    """Download a file by streming chunks of 1Mb.

    If the file already exists in local, it does not download it.
    """
    local_path = get_download_path(url, external_id, local_dir_path)
    local_path.parent.mkdir(parents=True, exist_ok=True)
    if not local_path.is_file():
        with requests.get(url, stream=True, timeout=20) as response:
            response.raise_for_status()
            with open(local_path, "wb") as file:
                for chunk in response.iter_content(chunk_size=1024 * 1024):
                    file.write(chunk)
    return str(local_path)


def assert_required_fields_existence(assets: List[Dict]) -> None:
    """Check if all fields are available to download assets."""
    required_fields = ["content", "externalId"]
    for field in required_fields:
        if field not in assets[0].keys():
            raise MissingPropertyError(
                f"The asset does not have the {field} field. Please add it to the fields when"
                " querying assets."
            )



"""Types specific to import."""
from typing import Dict, List, NewType, Optional

from pydantic import BaseModel, Extra, StrictInt, StrictStr, validator
from typing_extensions import Literal

Classes = NewType("Classes", Dict[int, str])
LabelFormat = Literal["yolo_v4", "yolo_v5", "yolo_v7", "kili", "raw"]


class ClientInputLabelData(BaseModel, extra=Extra.forbid):
    """Data about a label to append, given by client in client-side function."""

    asset_id: Optional[StrictStr]
    asset_external_id: Optional[StrictStr]
    json_response: dict
    author_id: Optional[StrictStr]
    seconds_to_label: Optional[StrictInt]


class _ClientInputLabelsValidator(BaseModel, extra=Extra.forbid):
    """Validates the data about a label to append."""

    labels: List[Dict]

    @validator("labels", each_item=True, pre=True)
    def label_validator(cls, label):  # pylint: disable=no-self-argument
        """Validate the data of one label."""
        if label.get("asset_external_id") is None and label.get("asset_id") is None:
            raise ValueError("You must either provide the asset_id or external_id")
        return ClientInputLabelData(**label)

    @validator("labels")
    def all_labels_use_the_same_asset_identifier(cls, labels):  # pylint: disable=no-self-argument
        """Validate the data of one label."""
        if not all(label.get("asset_id") for label in labels) and not all(
            label.get("asset_external_id") for label in labels
        ):
            raise ValueError(
                "Please use the same asset identifier for all labels: either only asset_id or only"
                " asset_external_id"
            )
        return labels



"""Set of common functions used by different export formats."""
import warnings
from typing import Dict, List, Optional

from kili.core.graphql import QueryOptions
from kili.core.graphql.operations.asset.queries import AssetQuery, AssetWhere
from kili.core.helpers import validate_category_search_query
from kili.entrypoints.queries.asset.media_downloader import get_download_assets_function
from kili.services.export.types import ExportType

DEFAULT_FIELDS = [
    "id",
    "content",
    "externalId",
    "jsonMetadata",
    "labels.author.id",
    "labels.author.email",
    "labels.author.firstname",
    "labels.author.lastname",
    "labels.jsonResponse",
    "labels.createdAt",
    "labels.isLatestLabelForUser",
    "labels.labelType",
    "labels.modelName",
]
LATEST_LABEL_FIELDS = [
    "id",
    "content",
    "externalId",
    "jsonContent",
    "jsonMetadata",
    "latestLabel.author.id",
    "latestLabel.author.email",
    "latestLabel.jsonResponse",
    "latestLabel.author.firstname",
    "latestLabel.author.lastname",
    "latestLabel.createdAt",
    "latestLabel.isLatestLabelForUser",
    "latestLabel.labelType",
    "latestLabel.modelName",
]


def attach_name_to_assets_labels_author(assets: List[Dict], export_type: ExportType):
    """Adds `name` field for author, by concatenating his/her first and last name."""
    for asset in assets:
        if export_type == "latest":
            latest_label = asset["latestLabel"]
            if latest_label:
                firstname = latest_label["author"]["firstname"]
                lastname = latest_label["author"]["lastname"]
                latest_label["author"]["name"] = f"{firstname} {lastname}"
            continue
        for label in asset.get("labels", []):
            firstname = label["author"]["firstname"]
            lastname = label["author"]["lastname"]
            label["author"]["name"] = f"{firstname} {lastname}"


THRESHOLD_WARN_MANY_ASSETS = 1000


# pylint: disable=too-many-arguments, too-many-locals, missing-type-doc
def fetch_assets(
    kili,
    project_id: str,
    asset_ids: Optional[List[str]],
    export_type: ExportType,
    label_type_in: Optional[List[str]] = None,
    disable_tqdm: bool = False,
    download_media: bool = False,
    local_media_dir: Optional[str] = None,
    asset_filter_kwargs: Optional[Dict[str, object]] = None,
) -> List[Dict]:
    """Fetches assets.

    Fetches assets where ID are in asset_ids if the list has more than one element,
    else all the assets of the project. If download media is passed, the media are
    downloaded into the `$HOME/.cache` folder.

    Args:
        kili: Kili instance
        project_id: project id
        asset_ids: list of asset IDs
        export_type: type of export (latest label or all labels)
        label_type_in: types of label to fetch (default, reviewed, ...)
        disable_tqdm: tell to disable tqdm
        download_media: tell to download the media in the cache folder.
        local_media_dir: Directory where the media are downloaded if `download_media` is True.
        asset_filter_kwargs: Optional dictionary of arguments to filter the assets to export.

    Returns:
        List of fetched assets.
    """
    fields = get_fields_to_fetch(export_type)

    asset_filter_kwargs = asset_filter_kwargs or {}
    asset_where_params = {
        "project_id": project_id,
        "label_type_in": label_type_in,
        "consensus_mark_gte": asset_filter_kwargs.pop("consensus_mark_gte", None),
        "consensus_mark_lte": asset_filter_kwargs.pop("consensus_mark_lte", None),
        "external_id_strictly_in": asset_filter_kwargs.pop(
            "external_id_strictly_in", None
        ) or asset_filter_kwargs.pop("external_id_contains", None),
        "external_id_in": asset_filter_kwargs.pop("external_id_in", None),
        "honeypot_mark_gte": asset_filter_kwargs.pop("honeypot_mark_gte", None),
        "honeypot_mark_lte": asset_filter_kwargs.pop("honeypot_mark_lte", None),
        "label_author_in": asset_filter_kwargs.pop("label_author_in", None),
        "label_reviewer_in": asset_filter_kwargs.pop("label_reviewer_in", None),
        "skipped": asset_filter_kwargs.pop("skipped", None),
        "status_in": asset_filter_kwargs.pop("status_in", None),
        "label_category_search": asset_filter_kwargs.pop("label_category_search", None),
        "created_at_gte": asset_filter_kwargs.pop("created_at_gte", None),
        "created_at_lte": asset_filter_kwargs.pop("created_at_lte", None),
        "issue_type": asset_filter_kwargs.pop("issue_type", None),
        "issue_status": asset_filter_kwargs.pop("issue_status", None),
        "inference_mark_gte": asset_filter_kwargs.pop("inference_mark_gte", None),
        "inference_mark_lte": asset_filter_kwargs.pop("inference_mark_lte", None),
        "metadata_where": asset_filter_kwargs.pop("metadata_where", None),
    }

    if asset_filter_kwargs:
        raise NameError(f"Unknown asset filter arguments: {list(asset_filter_kwargs.keys())}")

    if asset_where_params.get("label_category_search"):
        validate_category_search_query(asset_where_params["label_category_search"])  # type: ignore

    if asset_ids is not None and len(asset_ids) > 0:
        asset_where_params["asset_id_in"] = asset_ids

    where = AssetWhere(**asset_where_params)

    if download_media:
        count = AssetQuery(kili.graphql_client).count(where)
        if count > THRESHOLD_WARN_MANY_ASSETS:
            warnings.warn(
                (
                    f"Downloading many assets ({count}). This might take a while. Consider"
                    " disabling assets download in the options."
                ),
                stacklevel=3,
            )

    options = QueryOptions(disable_tqdm=disable_tqdm)
    post_call_function, fields = get_download_assets_function(
        kili, download_media, fields, project_id, local_media_dir
    )
    assets = list(AssetQuery(kili.graphql_client)(where, fields, options, post_call_function))
    attach_name_to_assets_labels_author(assets, export_type)
    return assets


def get_fields_to_fetch(export_type: ExportType):
    """Returns the fields to fetch depending on the export type."""
    if export_type == "latest":
        return LATEST_LABEL_FIELDS
    return DEFAULT_FIELDS



"""CLI's project copy subcommand."""

from typing import Optional

import click

from kili import services
from kili.entrypoints.cli.common_args import Options
from kili.entrypoints.cli.helpers import get_kili_client


@click.command(name="copy")
@Options.api_key
@Options.endpoint
@click.argument("from_project_id", type=str, required=True)
@click.option("--title", type=str, required=False, help="New project title.")
@click.option("--description", type=str, required=False, help="New project description.")
@click.option(
    "--with-json-interface/--without-json-interface",
    required=False,
    default=True,
    help="Copy json interface.",
)
@click.option(
    "--with-quality-settings/--without-quality-settings",
    required=False,
    default=True,
    help="Copy quality settings.",
)
@click.option(
    "--with-members/--without-members", required=False, default=True, help="Copy members."
)
@click.option("--with-assets/--without-assets", required=False, default=False, help="Copy assets.")
@click.option("--with-labels/--without-labels", required=False, default=False, help="Copy labels.")
# pylint: disable=too-many-arguments
def copy_project(
    api_key: Optional[str],
    endpoint: Optional[str],
    from_project_id: str,
    title: Optional[str] = None,
    description: Optional[str] = None,
    with_json_interface: bool = True,
    with_quality_settings: bool = True,
    with_members: bool = True,
    with_assets: bool = False,
    with_labels: bool = False,
) -> None:
    """Copy an existing Kili project.

    The copy can include or not the json interface, quality settings, members, assets and labels of
    the source project.

    By default, only the json interface, quality settings and project members are copied.

    If no `title` is provided, the source project title will be used.
    If no description is provided, the description will be set to an empty string.

    Returns the new project id and title once the copy is finished.

    \b
    !!! Examples
        Copy a project and set a new title and new description:
        ```
        kili project copy clbqn56b331234567890l41c0 \\
            --title "New project title" \\
            --description "New project description"
        ```
        Copy the json interface but not the members:
        ```
        kili project copy clbqn56b331234567890l41c0 \\
            --with-json-interface \\
            --without-members
        ```
    """
    kili = get_kili_client(api_key=api_key, api_endpoint=endpoint)
    new_proj_id = kili.copy_project(
        from_project_id=from_project_id,
        title=title,
        description=description,
        copy_json_interface=with_json_interface,
        copy_quality_settings=with_quality_settings,
        copy_members=with_members,
        copy_assets=with_assets,
        copy_labels=with_labels,
    )
    title = services.get_project_field(kili, new_proj_id, "title")
    print(f'Project copied successfully. New project id: "{new_proj_id}", with title: "{title}"')



"""GraphQL Queries of Projects."""


from typing import Optional

from kili.core.graphql import BaseQueryWhere, GraphQLQuery


class ProjectWhere(BaseQueryWhere):
    """Tuple to be passed to the ProjectQuery to restrict query."""

    # pylint: disable=too-many-arguments

    def __init__(
        self,
        project_id: Optional[str] = None,
        search_query: Optional[str] = None,
        should_relaunch_kpi_computation: Optional[bool] = None,
        updated_at_gte: Optional[str] = None,
        updated_at_lte: Optional[str] = None,
        archived: Optional[bool] = None,
    ):
        self.project_id = project_id
        self.search_query = search_query
        self.should_relaunch_kpi_computation = should_relaunch_kpi_computation
        self.updated_at_gte = updated_at_gte
        self.updated_at_lte = updated_at_lte
        self.archived = archived
        super().__init__()

    def graphql_where_builder(self):
        """Build the GraphQL Where payload sent in the resolver from the SDK ProjectWhere."""
        ret = {
            "id": self.project_id,
            "searchQuery": self.search_query,
            "shouldRelaunchKpiComputation": self.should_relaunch_kpi_computation,
            "updatedAtGte": self.updated_at_gte,
            "updatedAtLte": self.updated_at_lte,
        }
        if self.archived is not None:
            ret["archived"] = self.archived
        return ret


class ProjectQuery(GraphQLQuery):
    """Project query."""

    @staticmethod
    def query(fragment):
        """Return the GraphQL projects query."""
        return f"""
        query projects($where: ProjectWhere!, $first: PageSize!, $skip: Int!) {{
            data: projects(where: $where, first: $first, skip: $skip) {{
                {fragment}
            }}
        }}
        """

    COUNT_QUERY = """
    query countProjects($where: ProjectWhere!) {
        data: countProjects(where: $where)
    }
    """



"""Issue mutations."""

from itertools import repeat
from typing import Dict, List, Optional

from typeguard import typechecked
from typing_extensions import Literal

from kili.core.graphql import QueryOptions
from kili.core.graphql.graphql_client import GraphQLClient
from kili.core.graphql.operations.label.queries import LabelQuery, LabelWhere
from kili.core.helpers import deprecate, format_result
from kili.entrypoints.mutations.asset.helpers import get_asset_ids_or_throw_error
from kili.services.helpers import assert_all_arrays_have_same_size
from kili.utils.logcontext import for_all_methods, log_call

from .helpers import get_issue_numbers, get_labels_asset_ids_map
from .queries import GQL_CREATE_ISSUES


@for_all_methods(log_call, exclude=["__init__"])
class MutationsIssue:
    """Set of Issue mutations."""

    graphql_client: GraphQLClient

    # pylint: disable=too-many-arguments
    @deprecate(
        msg=(
            "append_to_issues is deprecated. Please use `create_issues` or `create_questions`"
            " instead. These new methods allow to import several issues or questions at the same"
            " time and provide better performances."
        )
    )
    @typechecked
    def append_to_issues(
        self,
        label_id: str,
        project_id: str,
        object_mid: Optional[str] = None,
        text: Optional[str] = None,
        type_: Literal["ISSUE", "QUESTION"] = "ISSUE",
    ) -> Dict:
        """Create an issue.

        !!! danger "Deprecated"
            append_to_issues is deprecated.
            Please use `create_issues` or `create_questions` instead.
            These new methods allow to import several issues or questions at the same time
            and provide better performances.

        Args:
            label_id: Id of the label to add an issue to
            object_mid: Mid of the object in the label to associate the issue to
            type_: type of the issue to add. Can be either "ISSUE" or "QUESTION"
            text: If given, write a comment related to the issue
            project_id: Id of the project

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.
        """
        issue_number = get_issue_numbers(self, project_id, type_, 1)[0]
        try:
            options = QueryOptions(disable_tqdm=True)
            where = LabelWhere(
                project_id=project_id,
                label_id=label_id,
            )
            asset_id: str = list(
                LabelQuery(self.graphql_client)(where=where, fields=["labelOf.id"], options=options)
            )[0]["labelOf"]["id"]
        except:
            # pylint: disable=raise-missing-from
            raise ValueError(
                f"Label ID {label_id} does not exist in the project of ID {project_id}"
            )
        variables = {
            "issues": [
                {
                    "issueNumber": issue_number,
                    "labelID": label_id,
                    "objectMid": object_mid,
                    "type": type_,
                    "assetId": asset_id,
                    "text": text,
                }
            ],
            "where": {"id": asset_id},
        }

        result = self.graphql_client.execute(GQL_CREATE_ISSUES, variables)
        return format_result("data", result)[0]

    @typechecked
    def create_issues(
        self,
        project_id: str,
        label_id_array: List[str],
        object_mid_array: Optional[List[Optional[str]]] = None,
        text_array: Optional[List[Optional[str]]] = None,
    ) -> List[Dict]:
        """Create an issue.

        Args:
            project_id: Id of the project.
            label_id_array: List of Ids of the labels to add an issue to.
            object_mid_array: List of mids of the objects in the labels to associate the issues to.
            text_array: List of texts to associate to the issues.

        Returns:
            A list of dictionary with the `id` key of the created issues.
        """
        assert_all_arrays_have_same_size([label_id_array, object_mid_array, text_array])
        issue_number_array = get_issue_numbers(self, project_id, "ISSUE", len(label_id_array))
        label_asset_ids_map = get_labels_asset_ids_map(self, project_id, label_id_array)
        variables = {
            "issues": [
                {
                    "issueNumber": issue_number,
                    "labelID": label_id,
                    "objectMid": object_mid,
                    "type": "ISSUE",
                    "assetId": label_asset_ids_map[label_id],
                    "text": text,
                }
                for (issue_number, label_id, object_mid, text) in zip(
                    issue_number_array,
                    label_id_array,
                    object_mid_array or repeat(None),
                    text_array or repeat(None),
                )
            ],
            "where": {"idIn": list(label_asset_ids_map.values())},
        }

        result = self.graphql_client.execute(GQL_CREATE_ISSUES, variables)
        return format_result("data", result)

    @typechecked
    def create_questions(
        self,
        project_id: str,
        text_array: List[Optional[str]],
        asset_id_array: Optional[List[str]] = None,
        asset_external_id_array: Optional[List[str]] = None,
    ) -> List[Dict]:
        # pylint:disable=line-too-long
        """Create questions.

        Args:
            project_id: Id of the project.
            text_array: List of question strings.
            asset_id_array: List of the assets to add the questions to.
            asset_external_id_array: List of the assets to add the questions to. Used if `asset_id_array` is not given.

        Returns:
            A list of dictionary with the `id` key of the created questions.
        """
        assert_all_arrays_have_same_size([text_array, asset_id_array])
        issue_number_array = get_issue_numbers(self, project_id, "QUESTION", len(text_array))
        asset_id_array = get_asset_ids_or_throw_error(
            self, asset_id_array, asset_external_id_array, project_id
        )
        variables = {
            "issues": [
                {"issueNumber": issue_number, "type": "QUESTION", "assetId": asset_id, "text": text}
                for (asset_id, text, issue_number) in zip(
                    asset_id_array, text_array, issue_number_array
                )
            ],
            "where": {"idIn": asset_id_array},
        }

        result = self.graphql_client.execute(GQL_CREATE_ISSUES, variables)
        return format_result("data", result)



"""Module for parsing labels returned by kili.labels()."""

from copy import deepcopy
from typing import Dict, Generator, Iterable, List, overload

from kili.services.label_data_parsing import json_response as json_response_module
from kili.services.label_data_parsing.types import InputType, Project


class ParsedLabel(Dict):
    """Class that represents a parsed label."""

    def __init__(self, label: Dict, json_interface: Dict, input_type: InputType) -> None:
        # pylint: disable=line-too-long
        """Class that represents a parsed label.

        The class behaves like a dict but adds the attribute `.jobs`.

        The original input label passed to this class is not modified.

        Args:
            label: Label to parse.
            json_interface: Json interface of the project.
            input_type: Type of assets of the project.

        !!! Example
            ```python
            from kili.utils.labels.parsing import ParsedLabel

            my_label = kili.labels("project_id")[0]  # my_label is a dict

            my_parsed_label = ParsedLabel(my_label, json_interface, input_type)  # ParsedLabel object

            # Access the job "JOB_0" data through the attribute ".jobs":
            print(my_parsed_label.jobs["JOB_0"])
            ```

        !!! info
            More information about the label parsing can be found in this [tutorial](https://python-sdk-docs.kili-technology.com/latest/sdk/tutorials/label_parsing/).
        """
        label_copy = deepcopy(label)
        json_response = label_copy.pop("jsonResponse", {})

        super().__init__(label_copy)

        project_info = Project(inputType=input_type, jsonInterface=json_interface["jobs"])

        self.jobs = json_response_module.ParsedJobs(
            project_info=project_info, json_response=json_response
        )

    def to_dict(self) -> Dict:
        """Returns a copy of the parsed label as a dict.

        !!! Example
            ```python
            my_parsed_label = ParsedLabel(my_dict_label, json_interface, input_type)

            # Convert back to native Python dictionary
            my_label_as_dict = label.to_dict()

            assert isinstance(my_label_as_dict, dict)  # True
            ```
        """
        ret = {k: deepcopy(v) for k, v in self.items() if k != "jsonResponse"}
        ret["jsonResponse"] = self.json_response
        return ret

    def __repr__(self) -> str:
        """Returns the representation of the object."""
        return repr(self.to_dict())

    def __str__(self) -> str:
        """Returns the string representation of the object."""
        return str(self.to_dict())

    @property
    def json_response(self) -> Dict:
        """Returns a copy of the json response of the parsed label."""
        return self.jobs.to_dict()


@overload
def parse_labels(
    labels: List[Dict], json_interface: Dict, input_type: InputType
) -> List[ParsedLabel]:
    ...


@overload
def parse_labels(
    labels: Generator[Dict, None, None], json_interface: Dict, input_type: InputType
) -> Generator[ParsedLabel, None, None]:
    ...


def parse_labels(
    labels: Iterable[Dict], json_interface: Dict, input_type: InputType
) -> Iterable[ParsedLabel]:
    """Parse labels returned by kili.labels().

    Args:
        labels: List or generator of labels from kili.labels().
        json_interface: Json interface of the project.
        input_type: Type of assets of the project.

    Returns:
        Parsed labels.
    """
    gen = (
        ParsedLabel(label=label, json_interface=json_interface, input_type=input_type)
        for label in labels
    )
    return gen if isinstance(labels, Generator) else list(gen)



"""Assets related mutations."""

GQL_APPEND_MANY_TO_DATASET = """
mutation appendManyToDataset(
    $data: AppendManyToDatasetData!,
    $where: ProjectWhere!
  ) {
  data: appendManyToDataset(
    data: $data,
    where: $where
  ) {
    id
  }
}
"""

GQL_APPEND_MANY_FRAMES_TO_DATASET = """
mutation(
    $data: AppendManyFramesToDatasetAsynchronouslyData!,
    $where: ProjectWhere!
  ) {
  data: appendManyFramesToDatasetAsynchronously(
    data: $data,
    where: $where
  ) {
    id
  }
}
"""



"""Module for the "annotations" key parsing of a job response."""

import functools
from collections import defaultdict
from typing import Dict, Iterator, List, Optional, Sequence, Union

from typeguard import typechecked
from typing_extensions import Literal

from kili.services.label_data_parsing import category as category_module
from kili.services.label_data_parsing import json_response as json_response_module

from .bounding_poly import BoundingPolyList
from .decorators import for_all_properties
from .exceptions import AttributeNotCompatibleWithJobError, InvalidMutationError
from .types import NormalizedVertex, Project
from .utils import get_children_job_names


class _BaseAnnotation:
    """Class for parsing the "annotations" key of a job response.

    It is used as a base class for the common properties of all types of annotations.
    """

    def __init__(self, annotation_json: Dict, project_info: Project, job_name: str) -> None:
        """Class for Annotation parsing.

        Args:
            annotation_json: Dict of an annotation. It is the value of
                the "annotations" key of a job response.
            project_info: Information about the project.
            job_name: Name of the job.
        """
        self._json_data = annotation_json
        self._project_info = project_info
        self._job_name = job_name

        self._job_interface = project_info["jsonInterface"][job_name]  # type: ignore

        # cast lists to objects
        if "categories" in self._json_data and isinstance(self._json_data["categories"], List):
            self._json_data["categories"] = category_module.CategoryList(
                categories_list=self._json_data["categories"],
                project_info=self._project_info,
                job_name=self._job_name,
            )
        if "boundingPoly" in self._json_data and isinstance(self._json_data["boundingPoly"], List):
            self._json_data["boundingPoly"] = BoundingPolyList(
                bounding_poly_list=self._json_data["boundingPoly"],
                project_info=self._project_info,
                job_name=self._job_name,
                type_of_tool=self._json_data.get("type"),
            )
        if "points" in self._json_data and isinstance(self._json_data["points"], List):
            self._json_data["points"] = [
                PoseEstimationPointAnnotation(
                    project_info=self._project_info, job_name=self._job_name, annotation_json=point
                )
                for point in self._json_data["points"]
            ]
        if self._json_data.get("children"):
            self.children = self._json_data["children"]  # call children setter

    def __str__(self) -> str:
        return str(self._json_data)

    def __repr__(self) -> str:
        return repr(self._json_data)

    def as_dict(self) -> Dict:
        """Returns the parsed annotation as a dict."""
        ret = {
            k: v
            for k, v in self._json_data.items()
            if k not in ("categories", "children", "boundingPoly", "points")
        }
        if "categories" in self._json_data:
            ret["categories"] = (
                self._json_data["categories"]
                if isinstance(self._json_data["categories"], List)
                else self._json_data["categories"].as_list()
            )
        if "boundingPoly" in self._json_data:
            ret["boundingPoly"] = (
                self._json_data["boundingPoly"]
                if isinstance(self._json_data["boundingPoly"], List)
                else self._json_data["boundingPoly"].as_list()
            )
        if "points" in self._json_data:
            ret["points"] = [point.as_dict() for point in self._json_data["points"]]
        if "children" in self._json_data:
            ret["children"] = (
                self._json_data["children"].to_dict()
                if not isinstance(self._json_data["children"], Dict)
                else self._json_data["children"]
            )
        return ret

    @property
    def categories(self) -> "category_module.CategoryList":
        """Returns the list of categories of the annotation."""
        return self._json_data["categories"]

    @property
    def category(self) -> "category_module.Category":
        """Returns the category of the annotation if there is only one category.

        Else raises an error.
        """
        if self._job_interface["content"]["input"] not in ("radio", "singleDropdown"):
            raise AttributeNotCompatibleWithJobError("category")

        if "categories" not in self._json_data and not self._job_interface["required"]:
            return None  # type: ignore

        if len(self._json_data["categories"]) != 1:
            raise ValueError(f"Expected 1 category, got {self._json_data['categories']}")

        return self._json_data["categories"][0]

    @typechecked
    def add_category(self, name: str, confidence: Optional[int] = None) -> None:
        """Adds a category to an annotation job with categories."""
        if "categories" not in self._json_data:
            category_list = category_module.CategoryList(
                categories_list=[], project_info=self._project_info, job_name=self._job_name
            )
            category_list.add_category(name=name, confidence=confidence)
            self._json_data["categories"] = category_list
        else:
            self._json_data["categories"].add_category(name=name, confidence=confidence)

    @property
    def mid(self) -> str:
        """Returns the annotation unique identifier."""
        return self._json_data["mid"]

    @mid.setter
    @typechecked
    def mid(self, mid: str) -> None:
        """Sets the annotation unique identifier."""
        if len(mid) == 0:
            raise ValueError("mid must be non-empty.")
        self._json_data["mid"] = mid

    @property
    def children(self) -> "json_response_module.ParsedJobs":
        """Returns the parsed children jobs of the job."""
        return self._json_data["children"]

    @children.setter
    @typechecked
    def children(self, children: Dict) -> None:
        """Sets the children jobs of the annotation job."""
        job_names_to_parse = get_children_job_names(
            json_interface=self._project_info["jsonInterface"],
            job_interface=self._job_interface,  # type: ignore
        )
        parsed_children_job = json_response_module.ParsedJobs(
            project_info=self._project_info,
            json_response=children,
            job_names_to_parse=job_names_to_parse,
        )
        self._json_data["children"] = parsed_children_job


class _BaseNamedEntityRecognitionAnnotation(_BaseAnnotation):
    """Base class for parsing the "content" key of an annotation for NER job response.

    Either simple NER or NER in PDFs.
    """

    @property
    def content(self) -> str:
        """Returns the content of the annotation."""
        return self._json_data["content"]

    @content.setter
    @typechecked
    def content(self, content: str) -> None:
        """Sets the content of the annotation."""
        self._json_data["content"] = content


class EntityAnnotation(_BaseNamedEntityRecognitionAnnotation):
    """Class for parsing the "annotations" key of a job response for named entities recognition."""

    @staticmethod
    def _get_compatible_ml_task() -> Literal["NAMED_ENTITIES_RECOGNITION"]:
        return "NAMED_ENTITIES_RECOGNITION"

    @property
    def begin_offset(self) -> int:
        """Returns the begin offset of the annotation."""
        return self._json_data["beginOffset"]

    @begin_offset.setter
    @typechecked
    def begin_offset(self, begin_offset: int) -> None:
        """Sets the begin offset of the annotation."""
        if begin_offset < 0:
            raise ValueError(f"begin_offset must be positive, got {begin_offset}")
        self._json_data["beginOffset"] = begin_offset

    @property
    def end_offset(self) -> int:
        """Returns the end offset of the annotation."""
        return self._json_data["endOffset"]

    @end_offset.setter
    @typechecked
    def end_offset(self, end_offset: int) -> None:
        """Sets the end offset of the annotation."""
        if end_offset < 0:
            raise ValueError(f"end_offset must be positive, got {end_offset}")
        self._json_data["endOffset"] = end_offset


class _BaseAnnotationWithTool(_BaseAnnotation):
    """Base class for annotations with a "type" key (tool used to create the annotation)."""

    @property
    def type(self) -> Literal["rectangle", "polygon", "semantic", "marker", "vector", "polyline"]:
        """Returns the tool of the annotation.

        One of "rectangle", "polygon", or "semantic" for 2D annotations.

        One of "marker", "vector", "polyline" for 1D annotations.
        """
        return self._json_data["type"]


class PointAnnotation(_BaseAnnotationWithTool):
    """Class for parsing the "annotations" key of a job response for 1D object detection jobs."""

    @staticmethod
    def _get_compatible_ml_task() -> Literal["OBJECT_DETECTION"]:
        return "OBJECT_DETECTION"

    @staticmethod
    def _get_compatible_type_of_tools() -> Sequence[Literal["marker"]]:
        return ("marker",)

    @property
    def point(self) -> NormalizedVertex:
        """Returns the point of a point detection job."""
        return self._json_data["point"]

    @point.setter
    @typechecked
    def point(self, point: NormalizedVertex) -> None:
        """Sets the point of a point detection job."""
        self._json_data["point"] = point


class PolyLineAnnotation(_BaseAnnotationWithTool):
    """Class for parsing the "annotations" key of a job response for 1D object detection jobs."""

    @staticmethod
    def _get_compatible_ml_task() -> Literal["OBJECT_DETECTION"]:
        return "OBJECT_DETECTION"

    @staticmethod
    def _get_compatible_type_of_tools() -> Sequence[Literal["vector", "polyline"]]:
        return ("vector", "polyline")

    @property
    def polyline(self) -> List[NormalizedVertex]:
        """Returns the polyline of a polyline detection job."""
        return self._json_data["polyline"]

    @polyline.setter
    @typechecked
    def polyline(self, polyline: List[NormalizedVertex]) -> None:
        """Sets the polyline of a polyline detection job."""
        self._json_data["polyline"] = polyline


class _BaseAnnotationWithBoundingPoly(_BaseAnnotation):
    """Base class for annotations with a "boundingPoly" key."""

    @property
    def bounding_poly(self) -> BoundingPolyList:
        """Returns the polygon of the object contour."""
        return self._json_data["boundingPoly"]


# pylint: disable=line-too-long
class EntityInPdfAnnotation(_BaseNamedEntityRecognitionAnnotation, _BaseAnnotationWithBoundingPoly):
    """Class for parsing the "annotations" key of a job response for named entities recognition in PDFs."""

    @staticmethod
    def _get_compatible_ml_task() -> Literal["NAMED_ENTITIES_RECOGNITION"]:
        return "NAMED_ENTITIES_RECOGNITION"

    @property
    def annotations(self) -> "AnnotationList":
        """Return the tist of positions of the annotation.

        For NER, when an annotation spans multiple lines, there will be multiple polys and a single boundingPoly.
        """
        return AnnotationList(
            job_name=self._job_name,
            project_info=self._project_info,
            annotations_list=self._json_data["annotations"],
        )

    @property
    def polys(
        self,
    ) -> List[Dict[Literal["normalizedVertices"], List[List[NormalizedVertex]]]]:
        """Return the coordinates from the different rectangles in the annotation.

        An annotation can have several rectangles (for example if the annotation covers more than one line).
        """
        return self._json_data["polys"]

    @property
    def page_number_array(self) -> List[int]:
        """Return the pages where the annotation appears."""
        return self._json_data["pageNumberArray"]


class _Base2DAnnotation(_BaseAnnotationWithTool, _BaseAnnotationWithBoundingPoly):
    """Base class for 2D annotations."""

    @staticmethod
    def _get_compatible_ml_task() -> Literal["OBJECT_DETECTION"]:
        return "OBJECT_DETECTION"

    @staticmethod
    def _get_compatible_type_of_tools() -> (
        Sequence[Literal["rectangle", "polygon", "semantic", "polyline", "vector"]]
    ):
        return ("rectangle", "polygon", "semantic", "polyline", "vector")

    def add_bounding_poly(
        self,
        bounding_poly_dict: Dict[
            Literal["normalizedVertices"],
            Union[List[NormalizedVertex], List[List[NormalizedVertex]]],
        ],
    ) -> None:
        """Adds a bounding polygon to the boundingPoly list."""
        bounding_poly_list = (
            BoundingPolyList(
                bounding_poly_list=[],
                project_info=self._project_info,
                job_name=self._job_name,
                type_of_tool=self._json_data["type"],
            )
            if "boundingPoly" not in self._json_data
            else self._json_data["boundingPoly"]
        )
        bounding_poly_list.add_bounding_poly(bounding_poly_dict)
        self._json_data["boundingPoly"] = bounding_poly_list

    @property
    def score(self) -> Optional[int]:
        """Returns the score which is the confidence of the object detection.

        Available when a pre-annotation model is used.
        """
        return self._json_data.get("score")

    @score.setter
    @typechecked
    def score(self, score: int) -> None:
        """Sets the score of the annotation."""
        if not 0 <= score <= 100:
            raise ValueError(f"Score must be between 0 and 100, got {score}.")
        self._json_data["score"] = score


class BoundingPolyAnnotation(_Base2DAnnotation):
    """Class for parsing the "annotations" key of a job response for 2D object detection jobs."""


class VideoAnnotation(_Base2DAnnotation):
    """Class for parsing the "annotations" key of a job response for video object detection jobs."""

    @property
    def is_key_frame(self) -> bool:
        """Returns the value of isKeyFrame for a video job.

        This is a Boolean indicating if the timestamp or frame is used for interpolation.
        """
        try:
            return self._json_data["isKeyFrame"]
        except KeyError as err:
            raise AttributeNotCompatibleWithJobError("is_key_frame") from err


class PoseEstimationAnnotation(_BaseAnnotationWithTool):
    """Class for parsing the "annotations" key of a job response for pose estimation jobs."""

    @staticmethod
    def _get_compatible_ml_task() -> Literal["OBJECT_DETECTION"]:
        return "OBJECT_DETECTION"

    @staticmethod
    def _get_compatible_type_of_tools() -> Sequence[Literal["pose", "marker"]]:
        return ("pose", "marker")

    @property
    def kind(self) -> Literal["POSE_ESTIMATION"]:
        """Returns the job kind.

        In pose estimation jobs, this is always "POSE_ESTIMATION".
        """
        return self._json_data["kind"]

    @property
    def points(self) -> List["PoseEstimationPointAnnotation"]:
        """Returns the list of the points composing the object."""
        return self._json_data["points"]


class PoseEstimationPointAnnotation(PointAnnotation):
    """Class for parsing a point object of a job response for pose estimation jobs."""

    @property
    def point(self) -> Dict[Literal["x", "y"], float]:
        """Returns the coordinates of the point."""
        return self._json_data["point"]

    @property
    def code(self) -> str:
        """Returns the code identifier (unique for each point in an object)."""
        return self._json_data["code"]

    @property
    def name(self) -> str:
        """Returns the name of the point."""
        return self._json_data["name"]

    @property
    def job_name(self) -> str:
        """Returns the annotated point job's name."""
        return self._json_data["jobName"]


class EntityRelationAnnotation(_BaseAnnotation):
    """Class for parsing the "annotations" key of a job response for named entity relation jobs."""

    @staticmethod
    def _get_compatible_ml_task() -> Literal["NAMED_ENTITIES_RELATION"]:
        return "NAMED_ENTITIES_RELATION"

    @property
    def start_entities(self) -> List[Dict[Literal["mid"], str]]:
        """Returns the list of the start entities composing the relation."""
        return self._json_data["startEntities"]

    @start_entities.setter
    @typechecked
    def start_entities(self, start_entities: List[Dict[Literal["mid"], str]]) -> None:
        """Sets the list of the start entities composing the relation."""
        self._json_data["startEntities"] = start_entities

    @property
    def end_entities(self) -> List[Dict[Literal["mid"], str]]:
        """Returns the list of the end entities composing the relation."""
        return self._json_data["endEntities"]

    @end_entities.setter
    @typechecked
    def end_entities(self, end_entities: List[Dict[Literal["mid"], str]]) -> None:
        """Sets the list of the end entities composing the relation."""
        self._json_data["endEntities"] = end_entities


class ObjectRelationAnnotation(_BaseAnnotation):
    """Class for parsing the "annotations" key of a job response for object relation jobs."""

    @staticmethod
    def _get_compatible_ml_task() -> Literal["OBJECT_RELATION"]:
        return "OBJECT_RELATION"

    @property
    def start_objects(self) -> List[Dict[Literal["mid"], str]]:
        """Returns the list of the start objects composing the relation."""
        return self._json_data["startObjects"]

    @start_objects.setter
    @typechecked
    def start_objects(self, start_objects: List[Dict[Literal["mid"], str]]) -> None:
        """Sets the list of the start objects composing the relation."""
        self._json_data["startObjects"] = start_objects

    @property
    def end_objects(self) -> List[Dict[Literal["mid"], str]]:
        """Returns the list of the end objects composing the relation."""
        return self._json_data["endObjects"]

    @end_objects.setter
    @typechecked
    def end_objects(self, end_objects: List[Dict[Literal["mid"], str]]) -> None:
        """Sets the list of the end objects composing the relation."""
        self._json_data["endObjects"] = end_objects


def check_attribute_compatible_with_job(func):
    """Raises an error if the decorated method is not compatible with the job."""

    @functools.wraps(func)
    # pylint: disable=protected-access
    def wrapper(self, *args, **kwargs):
        attribute_name = func.__name__

        if attribute_name not in self._valid_attributes_for_ml_task[self._job_interface["mlTask"]]:
            raise AttributeNotCompatibleWithJobError(attribute_name)

        if (
            "type" in self._json_data
            and attribute_name not in self._valid_attributes_for_tool[self._json_data["type"]]
        ):
            raise AttributeNotCompatibleWithJobError(attribute_name)

        return func(self, *args, **kwargs)

    return wrapper


@for_all_properties(check_attribute_compatible_with_job)
# pylint: disable=too-many-ancestors
class Annotation(
    EntityAnnotation,
    PointAnnotation,
    PolyLineAnnotation,
    EntityInPdfAnnotation,
    BoundingPolyAnnotation,
    VideoAnnotation,
    PoseEstimationAnnotation,
    EntityRelationAnnotation,
    ObjectRelationAnnotation,
):
    """Class for parsing the "annotations" key of a job response.

    Contains all attributes that can be found in a job response.
    """

    def __init__(self, json_data: Dict, project_info: Project, job_name: str) -> None:
        """Initializes an Annotation object.

        This class is used to parse the "annotations" key of a job response.

        Args:
            json_data: The json data of the annotation.
            project_info: The project info object.
            job_name: The name of the job.
        """
        # dictionaries to store the valid attributes/properties for each mlTask and type of tool
        self._valid_attributes_for_ml_task = defaultdict(set)
        self._valid_attributes_for_tool = defaultdict(set)

        for parent_class in self.__class__.__bases__:
            parent_class_methods = dir(parent_class)
            parent_class_properties = [
                method
                for method in parent_class_methods
                if isinstance(getattr(parent_class, method), property)
            ]

            ml_task = parent_class._get_compatible_ml_task()  # type: ignore
            self._valid_attributes_for_ml_task[ml_task].update(parent_class_properties)  # type: ignore

            try:
                type_of_tools = parent_class._get_compatible_type_of_tools()  # type: ignore
            except AttributeError:
                continue

            for tool in type_of_tools:
                self._valid_attributes_for_tool[tool].update(parent_class_properties)  # type: ignore

        super().__init__(annotation_json=json_data, project_info=project_info, job_name=job_name)

    # all properties should be inherited from the base classes


class AnnotationList:
    """Class for the annotations list parsing."""

    def __init__(self, annotations_list: List[Dict], project_info: Project, job_name: str) -> None:
        """Class for the annotations list parsing.

        Args:
            annotations_list: List of dicts representing annotations.
            project_info: The project info object.
            job_name: The name of the job.
        """
        self._project_info = project_info
        self._job_name = job_name
        self._job_interface = self._project_info["jsonInterface"][self._job_name]  # type: ignore

        self._annotations_list: List[Annotation] = []

        for annotation_dict in annotations_list:
            self.add_annotation(annotation_dict)

    def _check_can_append_annotation(self, annotation: Annotation) -> None:
        # pylint: disable=protected-access
        if (
            "type" in annotation._json_data
            and annotation._json_data["type"] not in self._job_interface["tools"]
        ):
            annotation_tool_type = annotation._json_data["type"]
            job_interface_tools = self._job_interface["tools"]

            # pose estimation jobs are a special case
            # the job interface tools is "pose"
            # but the annotation type is "marker"
            if annotation_tool_type == "marker" and "pose" in job_interface_tools:
                return

            if annotation_tool_type not in job_interface_tools:
                raise InvalidMutationError(
                    f"Annotation of type '{annotation._json_data['type']}' cannot be added to this"
                    f" job with tools {self._job_interface['tools']}"
                )

    @typechecked
    def add_annotation(self, annotation_dict: Dict) -> None:
        """Adds an annotation object to the AnnotationList object."""
        annotation = Annotation(
            json_data=annotation_dict, project_info=self._project_info, job_name=self._job_name
        )
        self._check_can_append_annotation(annotation)
        self._annotations_list.append(annotation)

    @typechecked
    def __getitem__(self, index: int) -> Annotation:
        """Returns the annotation object at the given index."""
        return self._annotations_list[index]

    def __len__(self) -> int:
        """Returns the number of annotations."""
        return len(self._annotations_list)

    def __iter__(self) -> Iterator[Annotation]:
        """Returns an iterator over the annotations."""
        return iter(self._annotations_list)

    def __str__(self) -> str:
        """Returns the string representation of the annotations list."""
        return str(self.as_list())

    def __repr__(self) -> str:
        """Returns the string representation of the annotations list."""
        return repr(self.as_list())

    def as_list(self) -> List[Dict]:
        """Returns the list of categories as a list of dicts."""
        return [annotation.as_dict() for annotation in self._annotations_list]



"""GraphQL Client."""

import json
import logging
import os
import random
import string
import threading
import time
from datetime import datetime
from pathlib import Path
from typing import Any, Callable, Dict, Optional, Union
from urllib.parse import urlparse

import graphql
import requests
import websocket
from filelock import FileLock
from gql import Client, gql
from gql.transport import exceptions
from gql.transport.exceptions import TransportServerError
from gql.transport.requests import RequestsHTTPTransport
from gql.transport.requests import log as gql_requests_logger
from graphql import DocumentNode, print_schema
from tenacity import Retrying
from tenacity.retry import retry_if_exception_type
from tenacity.stop import stop_after_delay
from tenacity.wait import wait_exponential
from typing_extensions import LiteralString

from kili import __version__
from kili.core.graphql.clientnames import GraphQLClientName
from kili.exceptions import GraphQLError
from kili.utils.logcontext import LogContext

DEFAULT_GRAPHQL_SCHEMA_CACHE_DIR = Path.home() / ".cache" / "kili" / "graphql"


# pylint: disable=too-many-instance-attributes, too-few-public-methods
class GraphQLClient:
    """GraphQL client."""

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        endpoint: str,
        api_key: str,
        client_name: GraphQLClientName,
        verify: bool = True,
        enable_schema_caching: bool = True,
        graphql_schema_cache_dir: Optional[Union[str, Path]] = DEFAULT_GRAPHQL_SCHEMA_CACHE_DIR,
    ) -> None:
        """Initialize the GraphQL client.

        Args:
            endpoint: Kili API endpoint.
            api_key: Kili API key.
            client_name: Name of the client.
            verify: Whether to verify the SSL certificate.
            enable_schema_caching: Whether to cache the GraphQL schema on disk.
            graphql_schema_cache_dir: Directory where to cache the GraphQL schema.
        """
        self.endpoint = endpoint
        self.api_key = api_key
        self.client_name = client_name
        self.verify = verify
        self.enable_schema_caching = enable_schema_caching
        self.graphql_schema_cache_dir = (
            Path(graphql_schema_cache_dir) if graphql_schema_cache_dir else None
        )

        self.ws_endpoint = self.endpoint.replace("http", "ws")

        # mutex to avoid multiple threads sending queries to the backend at the same time
        # (which can happen if we use the same client in multiple threads)
        self._execute_lock = threading.Lock()

        gql_requests_logger.setLevel(logging.WARNING)

        self._gql_transport = RequestsHTTPTransport(
            url=endpoint,
            headers=self._get_headers(),
            cookies=None,
            auth=None,
            use_json=True,
            timeout=30,
            verify=verify,
            retries=20,
            method="POST",
        )

        if self.enable_schema_caching is True:
            if self.graphql_schema_cache_dir is None:
                raise ValueError(
                    "You must specify a cache directory if you want to enable schema caching."
                )

            self.graphql_schema_cache_dir.mkdir(parents=True, exist_ok=True)

            # mutex to avoid multiple processes operating in cache directory at the same time
            self._cache_dir_lock = FileLock(
                self.graphql_schema_cache_dir / "cache_dir.lock", timeout=15
            )

        self._gql_client = self._initizalize_graphql_client()

    def _get_headers(self) -> Dict[str, str]:
        """Get the headers."""
        return {
            "Authorization": f"X-API-Key: {self.api_key}",
            "Accept": "application/json",
            "Content-Type": "application/json",
            "apollographql-client-name": self.client_name.value,
            "apollographql-client-version": __version__,
        }

    def _get_introspection_args(self) -> Dict[str, bool]:
        """Get the introspection arguments."""
        return {
            "descriptions": True,  # descriptions for the schema, types, fields, and arguments
            "specified_by_url": False,  # https://spec.graphql.org/draft/#sec--specifiedBy
            "directive_is_repeatable": True,  # include repeatability of directives
            "schema_description": True,  # include schema description
            "input_value_deprecation": True,  # request deprecated input fields
        }

    def _initizalize_graphql_client(self) -> Client:
        """Initialize the GraphQL client."""
        if os.environ.get("KILI_SDK_SKIP_CHECKS", None) is not None:
            return Client(
                transport=self._gql_transport,
                fetch_schema_from_transport=False,
                introspection_args=self._get_introspection_args(),
            )

        if self.enable_schema_caching is False:
            return Client(
                transport=self._gql_transport,
                fetch_schema_from_transport=True,
                introspection_args=self._get_introspection_args(),
            )

        # In some cases (local development), we cannot get the kili version from the backend
        # and therefore we cannot determine the schema version, so we don't cache the schema
        graphql_schema_path = self._get_graphql_schema_path()
        if graphql_schema_path is None:
            return Client(
                transport=self._gql_transport,
                fetch_schema_from_transport=True,
                introspection_args=self._get_introspection_args(),
            )

        with self._cache_dir_lock:
            if not (graphql_schema_path.is_file() and graphql_schema_path.stat().st_size > 0):
                self._purge_graphql_schema_cache_dir()  # delete old schema files
                schema_str = self._get_graphql_schema_from_endpoint()
                self._cache_graphql_schema(graphql_schema_path, schema_str)

            else:
                schema_str = graphql_schema_path.read_text(encoding="utf-8")

        return Client(
            transport=self._gql_transport,
            schema=schema_str,
            introspection_args=self._get_introspection_args(),
        )

    def _get_graphql_schema_from_endpoint(self) -> str:
        """Get the GraphQL schema from the endpoint."""
        with Client(
            transport=self._gql_transport,
            fetch_schema_from_transport=True,
            introspection_args=self._get_introspection_args(),
        ) as session:
            schema_str = print_schema(session.client.schema)  # type: ignore

        return schema_str

    def _cache_graphql_schema(self, graphql_schema_path: Path, schema_str: str) -> None:
        """Cache the graphql schema on disk."""
        with self._cache_dir_lock:
            with graphql_schema_path.open("w", encoding="utf-8") as file:
                file.write(schema_str)
                file.flush()

    def _purge_graphql_schema_cache_dir(self) -> None:
        """Purge the schema cache directory."""
        if self.graphql_schema_cache_dir is None:
            return

        with self._cache_dir_lock:
            for file in self.graphql_schema_cache_dir.glob("*.graphql"):
                file.unlink()

    def _get_graphql_schema_path(self) -> Optional[Path]:
        """Get the path of the GraphQL schema.

        Will return None if we cannot get the schema version.
        """
        if self.graphql_schema_cache_dir is None:
            return None

        endpoint_netloc = urlparse(self.endpoint).netloc
        version = self._get_kili_app_version()
        if version is None:
            return None
        filename = f"{endpoint_netloc}_{version}.graphql"
        return self.graphql_schema_cache_dir / filename

    def _get_kili_app_version(self) -> Optional[str]:
        """Get the version of the Kili app server.

        Returns None if the version cannot be retrieved.
        """
        url = self.endpoint.replace("/graphql", "/version")
        response = requests.get(url, verify=self.verify, timeout=30)
        if response.status_code == 200 and '"version":' in response.text:
            response_json = response.json()
            version = response_json["version"]
            return version
        return None

    def execute(
        self, query: Union[str, DocumentNode], variables: Optional[Dict] = None, **kwargs
    ) -> Dict[str, Any]:
        """Execute a query.

        Args:
            query: the GraphQL query
            variables: the payload of the query
            kwargs: additional arguments to pass to the GraphQL client
        """

        def _execute(
            document: DocumentNode, variables: Optional[Dict] = None, **kwargs
        ) -> Dict[str, Any]:
            try:
                result = self._gql_client.execute(
                    document=document,
                    variable_values=variables,
                    extra_args={
                        "headers": {
                            **self._gql_transport.headers,  # type: ignore
                            **LogContext(),
                        }
                    },
                    **kwargs,
                )

            except graphql.GraphQLError as err:  # local validation error
                raise GraphQLError(error=err.message) from err
            except exceptions.TransportQueryError as err:  # graphql query refused by the backend
                raise GraphQLError(error=err.errors) from err

            return result

        document = query if isinstance(query, DocumentNode) else gql(query)

        try:
            with self._execute_lock:
                return _execute(document, variables, **kwargs)

        except GraphQLError as err:
            # if error is due do parsing or local validation of the query (graphql.GraphQLError)
            # we refresh the schema and retry once
            if isinstance(err.__cause__, graphql.GraphQLError):
                self._purge_graphql_schema_cache_dir()
                self._gql_client = self._initizalize_graphql_client()
                return _execute(document, variables, **kwargs)  # if it fails again, we crash here

            raise err

        except TransportServerError as err:
            # if backend returned a 401 error, we retry a few times
            if err.code == 401:
                for attempt in Retrying(
                    stop=stop_after_delay(30),
                    wait=wait_exponential(multiplier=1, min=2, max=10),
                    retry=retry_if_exception_type(TransportServerError),
                    reraise=True,
                ):
                    with attempt:
                        return _execute(document, variables, **kwargs)

            raise err


GQL_WS_SUBPROTOCOL = "graphql-ws"


class SubscriptionGraphQLClient:
    """A simple GraphQL client that works over Websocket as the transport
    protocol, instead of HTTP.

    This follows the Apollo protocol.
    https://github.com/apollographql/subscriptions-transport-ws/blob/master/PROTOCOL.md
    """

    # pylint: disable=too-many-instance-attributes, too-many-arguments

    def __init__(self, url):
        self.ws_url = url
        self._id = None
        self._paused = False
        self._connect()
        self._subscription_running = False
        self._st_id = None
        self.failed_connection_attempts = 0

    def _connect(self):
        """Handles the connection."""
        self._conn = websocket.create_connection(
            self.ws_url, on_message=self._on_message, subprotocols=[GQL_WS_SUBPROTOCOL]
        )
        self._created_at = datetime.now()
        self._conn.on_message = self._on_message  # type: ignore

    def _reconnect(self):
        """Handles the reconnection."""
        self._connect()
        self._subscription_running = True
        dt_string = datetime.now().strftime(r"%d/%m/%Y %H:%M:%S")
        print(f"{dt_string} reconnected")
        self.failed_connection_attempts = 0

    def _on_message(self, message):
        """Handles messages.

        Args:
            message : the message
        """
        data = json.loads(message)
        # skip keepalive messages
        if data["type"] != "ka":
            print(message)

    def _conn_init(self, headers=None, authorization=None):
        """Initializes the websocket connection.

        Args:
            headers : Headers are necessary for Kili API v1
            authorization : Headers are necessary for Kili API v2
        """
        payload = {
            "type": "connection_init",
            "payload": {"headers": headers, "Authorization": authorization},
        }
        self._conn.send(json.dumps(payload).encode("utf-8"))
        self._conn.recv()

    def _start(self, payload):
        """Handles start.

        Args:
            payload
        """
        _id = gen_id()
        frame = {"id": _id, "type": "start", "payload": payload}
        self._conn.send(json.dumps(frame).encode("utf-8"))
        return _id

    def _stop(self, _id):
        """Handles stop.

        Args:
        - _id: connection id
        """
        payload = {"id": _id, "type": "stop"}
        self._conn.send(json.dumps(payload).encode("utf-8"))
        return self._conn.recv()

    def query(self, query: str, variables: Optional[Dict] = None, headers: Optional[Dict] = None):
        """Sends a query.

        Args:
            query: the GraphQL query
            variables: the payload of the query
            headers: headers
        """
        self._conn_init(headers)
        payload = {"headers": headers, "query": query, "variables": variables}
        _id = self._start(payload)
        res = self._conn.recv()
        self._stop(_id)
        return res

    def prepare_subscribe(
        self,
        query: str,
        variables: Optional[Dict],
        headers: Optional[Dict],
        callback: Optional[Callable],
        authorization: Optional[str],
    ):
        """Prepares a subscription.

        Args:
            query: the GraphQL query
            variables: the payload of the query
            headers: headers
            callback: function executed after the subscription
            authorization: authorization header
        """
        self._conn_init(headers, authorization)
        payload = {"headers": headers, "query": query, "variables": variables}
        _cc = self._on_message if not callback else callback
        _id = self._start(payload)
        self._id = _id
        return _cc, _id

    def subscribe(
        self,
        query: str,
        variables: Optional[Dict] = None,
        headers: Optional[Dict] = None,
        callback: Optional[Callable] = None,
        authorization: Optional[str] = None,
    ):
        """Subscribes.

        Args:
            query: the GraphQL query
            variables: the payload of the query
            headers: headers
            callback: function executed after the subscription
            authorization: authorization header
        """
        _cc, _id = self.prepare_subscribe(query, variables, headers, callback, authorization)

        def subs(_cc, _id):
            max_reconnections = 10
            self._subscription_running = True
            while (
                self._subscription_running and self.failed_connection_attempts < max_reconnections
            ):
                try:
                    response = json.loads(self._conn.recv())
                    if response["type"] == "error" or response["type"] == "complete":
                        print(response)
                        self._stop_subscribe(_id)
                        break
                    if response["type"] != "ka" and not self._paused:
                        _cc(_id, response)  # type:ignore
                    time.sleep(1)
                except (
                    websocket._exceptions.WebSocketConnectionClosedException  # pylint: disable=protected-access
                ) as error:
                    self.failed_connection_attempts += 1
                    dt_string = datetime.now().strftime("%d/%m/%Y %H:%M:%S")
                    error_message = str(error)
                    print(f"{dt_string} Connection closed error : {error_message}")
                    print(
                        "Will try to reconnect"
                        f" {max_reconnections - self.failed_connection_attempts} times..."
                    )
                    self._reconnect()
                    _cc, _id = self.prepare_subscribe(
                        query, variables, headers, callback, authorization
                    )
                    continue
            print(f"Did not reconnect successfully after {max_reconnections} attempts")

        self._st_id = threading.Thread(target=subs, args=(_cc, _id))
        self._st_id.start()
        return _id

    def _stop_subscribe(self, _id):
        self._subscription_running = False
        self._stop(_id)

    def close(self):
        """Handles close."""
        self._conn.close()

    def pause(self):
        """Handles pause."""
        self._paused = True

    def unpause(self):
        """Handles unpause."""
        self._paused = False

    def get_lifetime(self):
        """Return the lifetime."""
        return (datetime.now() - self._created_at).seconds

    def reset_timeout(self):
        """Resets the timeout."""
        self._reconnect()


def gen_id(size: int = 6, chars: LiteralString = string.ascii_letters + string.digits) -> str:
    """Generate random alphanumeric id.

    Args:
        size: length of the id
        chars: chars used to generate the id
    """
    return "".join(random.choice(chars) for _ in range(size))



"""Utils to log calls."""
import functools
import platform
import uuid
from datetime import datetime, timezone
from typing import Callable, Dict, List

from kili import __version__
from kili.core.graphql.clientnames import GraphQLClientName


class Singleton(type):
    """Utility meta-class to guarantee single instance."""

    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


class LogContext(Dict[str, str], metaclass=Singleton):
    # pylint: disable=line-too-long
    """Dict-like singleton that holds the data for the log contex, to be passed to the request headers."""

    def __init__(self) -> None:
        self["kili-client-name"] = GraphQLClientName.SDK.value
        self["kili-client-version"] = __version__
        self["kili-client-language-name"] = "Python"
        self["kili-client-language-version"] = platform.python_version()
        self["kili-client-platform-version"] = platform.version()
        self["kili-client-platform-name"] = platform.system()


def for_all_methods(decorator: Callable, exclude: List[str]):
    """Class Decorator to decorate all the method with a decorator passed as argument."""

    def decorate(cls):
        for attr in cls.__dict__:
            if callable(getattr(cls, attr)) and attr not in exclude:
                setattr(cls, attr, decorator(getattr(cls, attr)))
        return cls

    return decorate


def log_call(func: Callable):
    """Decorator to add call info to the client logging context."""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        context = LogContext()
        context["kili-client-method-name"] = func.__name__
        context["kili-client-call-time"] = (
            datetime.now(timezone.utc).isoformat().replace("+00:00", "Z")
        )
        context["kili-client-call-uuid"] = str(uuid.uuid4())
        return func(*args, **kwargs)

    return wrapper



"""Class to upload a plugin."""

import ast
import time
from pathlib import Path
from typing import List, Optional, Tuple, Union
from zipfile import ZipFile

from kili.core.constants import (
    mime_extensions_for_py_scripts,
    mime_extensions_for_txt_files,
)
from kili.core.graphql.operations.plugin.mutations import (
    GQL_CREATE_PLUGIN,
    GQL_CREATE_PLUGIN_RUNNER,
    GQL_CREATE_WEBHOOK,
    GQL_GENERATE_UPDATE_URL,
    GQL_UPDATE_PLUGIN_RUNNER,
    GQL_UPDATE_WEBHOOK,
)
from kili.core.graphql.operations.plugin.queries import GQL_GET_PLUGIN_RUNNER_STATUS
from kili.core.helpers import format_result, get_data_type
from kili.services.plugins.tools import check_errors_plugin_upload
from kili.utils import bucket
from kili.utils.tempfile import TemporaryDirectory

from .exceptions import PluginCreationError
from .helpers import get_logger

NUMBER_TRIES_RUNNER_STATUS = 20

POSSIBLE_HANDLERS = {
    "on_submit": "onSubmit",
    "on_review": "onReview",
    "on_custom_interface_click": "onCustomInterfaceClick",
}


def check_file_mime_type(
    path: Path, compatible_mime_extensions: List[str], verbose: bool = True
) -> bool:
    # pylint: disable=line-too-long
    """Returns true if the mime type of the file corresponds to one of compatible_mime_extensions."""
    mime_type = get_data_type(path.as_posix())

    if not (compatible_mime_extensions and mime_type):
        return False

    correct_mime_type = mime_type in compatible_mime_extensions

    if verbose and not correct_mime_type:
        print(
            f"File mime type for {path} is {mime_type} and does not correspond"
            "to the type of the project. "
            f"File mime type should be one of {compatible_mime_extensions}"
        )
    return correct_mime_type


def check_file_is_py(path: Path, verbose: bool = True) -> bool:
    """Returns true if the mime type of the file corresponds to a python file."""
    return check_file_mime_type(path, mime_extensions_for_py_scripts, verbose)


def check_file_is_txt(path: Path, verbose: bool = True) -> bool:
    """Returns true if the mime type of the file corresponds to a .txt file."""
    return check_file_mime_type(path, mime_extensions_for_txt_files, verbose)


def check_file_contains_handler(path: Path) -> Tuple[bool, Optional[List[str]]]:
    """Return true if the file contain PluginHandler Class."""
    with open(path, encoding="utf-8") as file:
        module = ast.parse(file.read())
    for node in module.body:
        if isinstance(node, ast.ClassDef) and node.name == "PluginHandler":
            handlers = []
            for child in node.body:
                if isinstance(child, ast.FunctionDef) and child.name in POSSIBLE_HANDLERS:
                    handlers.append(POSSIBLE_HANDLERS[child.name])
            return True, handlers
    return False, None


class WebhookUploader:
    """Class to create a webhook."""

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        kili,
        webhook_url: str,
        plugin_name: str,
        header: Optional[str],
        verbose: bool,
        handler_types: Optional[List[str]],
    ) -> None:
        self.kili = kili
        self.webhook_url = webhook_url
        self.plugin_name = plugin_name or self.webhook_url
        self.header = header
        self.verbose = verbose
        self.handler_types = handler_types

    def create_webhook(self):
        """Create a webhook receiving Kili events."""
        variables = {
            "handlerTypes": self.handler_types,
            "pluginName": self.plugin_name,
            "webhookUrl": self.webhook_url,
            "header": self.header,
        }

        result = self.kili.graphql_client.execute(GQL_CREATE_WEBHOOK, variables)

        return format_result("data", result)

    def update_webhook(self):
        """Update a webhook receiving Kili events."""
        variables = {
            "handlerTypes": self.handler_types,
            "pluginName": self.plugin_name,
            "webhookUrl": self.webhook_url,
            "header": self.header,
        }

        result = self.kili.graphql_client.execute(GQL_UPDATE_WEBHOOK, variables)

        return format_result("data", result)


class PluginUploader:
    """Class to upload a plugin."""

    def __init__(self, kili, plugin_path: str, plugin_name: Optional[str], verbose: bool) -> None:
        self.kili = kili
        self.plugin_path = Path(plugin_path)

        if (not self.plugin_path.is_dir()) and (not self.plugin_path.is_file()):
            raise FileNotFoundError(
                f"The provided path '{plugin_path}' is neither a directory nor a file. The absolute"
                f" path is the following: '{self.plugin_path.absolute()}'"
            )

        if plugin_name:
            self.plugin_name = plugin_name
        else:
            self.plugin_name = self.plugin_path.name
        self.verbose = verbose
        self.handler_types = None

    def _retrieve_plugin_src(self) -> List[Path]:
        # pylint: disable=line-too-long
        """Retrieve script from plugin_path and execute it to prevent an upload with indentation errors."""
        if self.plugin_path.is_dir():
            file_path = self.plugin_path / "main.py"
            if not file_path.is_file():
                raise FileNotFoundError(
                    f"No main.py file in the provided folder: {self.plugin_path.absolute()}"
                )
            contains_handler, handler_types = check_file_contains_handler(file_path)
            if not contains_handler:
                raise ValueError("PluginHandler class is not present in your main.py file.")

            self.handler_types = handler_types

            file_paths = list(self.plugin_path.glob("**/*.py"))

            return file_paths

        file_path = self.plugin_path

        if not check_file_is_py(file_path, self.verbose):
            raise ValueError("Wrong file format.")

        contains_handler, handler_types = check_file_contains_handler(file_path)
        if not contains_handler:
            raise ValueError("PluginHandler class is not present in your plugin file.")

        self.handler_types = handler_types

        return [file_path]

    def _retrieve_requirements(self) -> Union[Path, None]:
        # pylint: disable=line-too-long
        """Retrieve script from file_path and execute it to prevent an upload with indentation errors."""
        if not self.plugin_path.is_dir():
            return None

        file_path = self.plugin_path / "requirements.txt"
        if not file_path.is_file():
            logger = get_logger()
            logger.info(
                f"No requirements.txt file in the provided folder: {self.plugin_path.absolute()}"
            )
            return None

        if not check_file_is_txt(file_path, self.verbose):
            raise ValueError("Wrong file format.")

        return file_path

    @staticmethod
    def _parse_script(script_path: Path):
        """Method to detect indentation and class errors in the script."""
        with script_path.open("r", encoding="utf-8") as file:
            source_code = file.read()

        # We execute the source code to prevent the upload of a file with SyntaxError
        compile(source_code, "<string>", "exec")

    @staticmethod
    def _upload_file(zip_path: Path, url: str):
        """Upload a file to a signed url and returns the url with the file_id."""
        bucket.upload_data_via_rest(url, zip_path.read_bytes(), "application/zip")

    def _retrieve_upload_url(self, is_updating_plugin: bool) -> str:
        """Retrieve an upload url from the backend."""
        variables = {"pluginName": self.plugin_name}

        if is_updating_plugin:
            result = self.kili.graphql_client.execute(GQL_GENERATE_UPDATE_URL, variables)
        else:
            result = self.kili.graphql_client.execute(GQL_CREATE_PLUGIN, variables)

        check_errors_plugin_upload(result, self.plugin_path, self.plugin_name)
        upload_url = format_result("data", result)
        return upload_url

    def _create_zip(self, tmp_directory: Path):
        """Create a zip file from python file and requirements.txt.

        (if user has defined a path to a requirements.txt)
        """
        file_paths = self._retrieve_plugin_src()

        for path in file_paths:
            self._parse_script(path)

        requirements_path = self._retrieve_requirements()

        zip_path = tmp_directory / "archive.zip"

        with ZipFile(zip_path, "w") as archive:
            if len(file_paths) == 1:
                archive.write(file_paths[0], "main.py")

            else:
                for path in file_paths:
                    archive.write(path, path.relative_to(self.plugin_path))

            if requirements_path:
                archive.write(requirements_path, requirements_path.relative_to(self.plugin_path))

        return zip_path

    def _upload_script(self, is_updating_plugin: bool):
        """Upload a script to Kili bucket."""
        with TemporaryDirectory() as tmp_directory:
            zip_path = self._create_zip(tmp_directory)

            upload_url = self._retrieve_upload_url(is_updating_plugin)

            self._upload_file(zip_path, upload_url)

    def _create_plugin_runner(self):
        """Create plugin's runner."""
        variables = {"pluginName": self.plugin_name, "handlerTypes": self.handler_types}

        result = self.kili.graphql_client.execute(GQL_CREATE_PLUGIN_RUNNER, variables)
        return format_result("data", result)

    def _check_plugin_runner_status(self, update=False):
        """Check the status of a plugin's runner until it is active."""
        logger = get_logger()

        action = "updated" if update else "created"

        logger.info(f"Plugin is being {action}... This should take approximately 3 minutes.")

        n_tries = 0
        status = None

        while n_tries < NUMBER_TRIES_RUNNER_STATUS:
            time.sleep(15)

            status = self.get_plugin_runner_status()

            if status in ("ACTIVE", "FAILED"):
                break

            logger.info(f"Status : {status}...")

            n_tries += 1

        if status == "DEPLOYING" and n_tries == 20:
            raise RuntimeError(
                "We could not check your plugin was deployed in time. Please check again the"
                " status of the plugin after some minutes with the command:"
                f' kili.get_plugin_status("{self.plugin_name}"). If the status is different than'
                " DEPLOYING or ACTIVE, please check your plugin's code and try to overwrite the"
                " plugin with a new version of the code (you can use kili.update_plugin() for"
                " that)."
            )

        if status != "ACTIVE":
            raise PluginCreationError(
                "There was some error during the creation of the plugin. Please check your"
                " plugin's code and try to overwrite the plugin with a new version of the code"
                " (you can use kili.update_plugin() for that). Hint: You can get build errors"
                f' using: kili.get_plugin_build_errors(plugin_name="{self.plugin_name}")'
            )

        message = f"Plugin {action} successfully"
        logger.info(message)

        return message

    def get_plugin_runner_status(self):
        """Get the status of a plugin's runner."""
        variables = {"name": self.plugin_name}

        result = self.kili.graphql_client.execute(GQL_GET_PLUGIN_RUNNER_STATUS, variables)

        return format_result("data", result)

    def create_plugin(self):
        """Create a plugin in Kili."""
        self._upload_script(False)

        self._create_plugin_runner()

        return self._check_plugin_runner_status()

    def _update_plugin_runner(self):
        """Update plugin's runner."""
        variables = {"pluginName": self.plugin_name, "handlerTypes": self.handler_types}

        result = self.kili.graphql_client.execute(GQL_UPDATE_PLUGIN_RUNNER, variables)
        return format_result("data", result)

    def update_plugin(self):
        """Update a plugin in Kili."""
        self._upload_script(True)

        self._update_plugin_runner()

        return self._check_plugin_runner_status(update=True)



"""This script permits to initialize the Kili Python SDK client."""
import os
import sys
import warnings
from datetime import datetime, timedelta
from typing import Dict, Optional

import requests

from kili import __version__
from kili.core.graphql import QueryOptions
from kili.core.graphql.graphql_client import GraphQLClient, GraphQLClientName
from kili.core.graphql.operations.api_key.queries import APIKeyQuery, APIKeyWhere
from kili.core.graphql.operations.user.queries import GQL_ME
from kili.core.helpers import format_result
from kili.entrypoints.mutations.asset import MutationsAsset
from kili.entrypoints.mutations.data_connection import MutationsDataConnection
from kili.entrypoints.mutations.issue import MutationsIssue
from kili.entrypoints.mutations.label import MutationsLabel
from kili.entrypoints.mutations.notification import MutationsNotification
from kili.entrypoints.mutations.plugins import MutationsPlugins
from kili.entrypoints.mutations.project import MutationsProject
from kili.entrypoints.mutations.project_version import MutationsProjectVersion
from kili.entrypoints.mutations.user import MutationsUser
from kili.entrypoints.queries.asset import QueriesAsset
from kili.entrypoints.queries.data_connection import QueriesDataConnection
from kili.entrypoints.queries.data_integration import QueriesDataIntegration
from kili.entrypoints.queries.issue import QueriesIssue
from kili.entrypoints.queries.label import QueriesLabel
from kili.entrypoints.queries.notification import QueriesNotification
from kili.entrypoints.queries.organization import QueriesOrganization
from kili.entrypoints.queries.plugins import QueriesPlugins
from kili.entrypoints.queries.project import QueriesProject
from kili.entrypoints.queries.project_user import QueriesProjectUser
from kili.entrypoints.queries.project_version import QueriesProjectVersion
from kili.entrypoints.queries.user import QueriesUser
from kili.entrypoints.subscriptions.label import SubscriptionsLabel
from kili.exceptions import AuthenticationFailed, UserNotFoundError
from kili.internal import KiliInternal

warnings.filterwarnings("default", module="kili", category=DeprecationWarning)


class Kili(  # pylint: disable=too-many-ancestors
    MutationsAsset,
    MutationsDataConnection,
    MutationsIssue,
    MutationsLabel,
    MutationsNotification,
    MutationsPlugins,
    MutationsProject,
    MutationsProjectVersion,
    MutationsUser,
    QueriesAsset,
    QueriesDataConnection,
    QueriesDataIntegration,
    QueriesIssue,
    QueriesLabel,
    QueriesNotification,
    QueriesOrganization,
    QueriesPlugins,
    QueriesProject,
    QueriesProjectUser,
    QueriesProjectVersion,
    QueriesUser,
    SubscriptionsLabel,
):
    """Kili Client."""

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        api_key: Optional[str] = None,
        api_endpoint: Optional[str] = None,
        verify: bool = True,
        client_name: GraphQLClientName = GraphQLClientName.SDK,
        graphql_client_params: Optional[Dict[str, object]] = None,
    ) -> None:
        """Initialize Kili client.

        Args:
            api_key: User API key generated
                from https://cloud.kili-technology.com/label/my-account/api-key.
                Default to `KILI_API_KEY` environment variable.
                If not passed, requires the `KILI_API_KEY` environment variable to be set.
            api_endpoint: Recipient of the HTTP operation.
                Default to `KILI_API_ENDPOINT` environment variable.
                If not passed, default to Kili SaaS:
                'https://cloud.kili-technology.com/api/label/v2/graphql'
            verify: Verify certificate. Set to False on local deployment without SSL.
            client_name: For internal use only.
                Define the name of the graphQL client whith which graphQL calls will be sent.
            graphql_client_params: Parameters to pass to the graphQL client.

        Returns:
            Instance of the Kili client.

        Examples:
            ```python
            from kili.client import Kili

            kili = Kili()

            kili.assets()  # list your assets
            kili.labels()  # list your labels
            kili.projects()  # list your projects
            ```
        """
        _deprecate_python_37()

        if api_key is None:
            api_key = os.getenv("KILI_API_KEY")

        if api_endpoint is None:
            api_endpoint = os.getenv(
                "KILI_API_ENDPOINT",
                "https://cloud.kili-technology.com/api/label/v2/graphql",
            )

        if api_key is None:
            raise AuthenticationFailed(api_key, api_endpoint)

        self.api_key = api_key
        self.api_endpoint = api_endpoint
        self.verify = verify
        self.client_name = client_name
        self.graphql_client_params = graphql_client_params

        skip_checks = os.environ.get("KILI_SDK_SKIP_CHECKS", None) is not None

        if not skip_checks:
            self._check_api_key_valid()

        self.graphql_client = GraphQLClient(
            endpoint=api_endpoint,
            api_key=api_key,
            client_name=client_name,
            verify=self.verify,
            **(graphql_client_params or {}),  # type: ignore
        )

        if not skip_checks:
            self._check_expiry_of_key_is_close()

        self.internal = KiliInternal(self)

    def _check_api_key_valid(self) -> None:
        """Check that the api_key provided is valid."""
        response = requests.post(
            url=self.api_endpoint,
            data='{"query":"{ me { id email } }"}',
            verify=self.verify,
            timeout=30,
            headers={
                "Authorization": f"X-API-Key: {self.api_key}",
                "Accept": "application/json",
                "Content-Type": "application/json",
                "apollographql-client-name": self.client_name.value,
                "apollographql-client-version": __version__,
            },
        )
        if response.status_code == 200 and "email" in response.text and "id" in response.text:
            return

        raise AuthenticationFailed(
            api_key=self.api_key,
            api_endpoint=self.api_endpoint,
            error_msg=(
                "Cannot check API key validity: status_code"
                f" {response.status_code}\n\n{response.text}"
            ),
        )

    def _check_expiry_of_key_is_close(self) -> None:
        """Check that the expiration date of the api_key is not too close."""
        warn_days = 30

        api_keys = APIKeyQuery(self.graphql_client)(
            fields=["expiryDate"],
            where=APIKeyWhere(api_key=self.api_key),
            options=QueryOptions(disable_tqdm=True),
        )

        key_expiry = datetime.strptime(next(api_keys)["expiryDate"], r"%Y-%m-%dT%H:%M:%S.%fZ")
        key_remaining_time = key_expiry - datetime.now()
        key_soon_deprecated = key_remaining_time < timedelta(days=warn_days)
        if key_soon_deprecated:
            message = f"""
                Your api key will be deprecated on {key_expiry:%Y-%m-%d}.
                You should generate a new one on My account > API KEY."""
            warnings.warn(message, UserWarning, stacklevel=2)

    def get_user(self) -> Dict:
        """Get the current user from the api_key provided."""
        result = self.graphql_client.execute(GQL_ME)
        user = format_result("data", result)
        if user is None or user["id"] is None or user["email"] is None:
            raise UserNotFoundError("No user attached to the API key was found")
        return user


def _deprecate_python_37() -> None:
    if sys.version_info < (3, 8):
        warnings.warn(
            (
                "Kili SDK will stop supporting Python 3.7 in July"
                " 2023. Please upgrade to Python 3.8 or higher."
            ),
            DeprecationWarning,
            stacklevel=1,
        )



"""Code specific to Azure blob storage."""

from pathlib import Path
from typing import Dict, List, Tuple
from urllib.parse import urlparse

from azure.storage.blob import BlobServiceClient


class AzureBucket:
    """Class for Azure blob storage buckets."""

    def __init__(self, sas_token: str, connection_url: str) -> None:
        """Initialize the Azure bucket."""
        self.sas_token = sas_token
        self.connection_url = connection_url

        self.storage_account, self.container_name = (
            self._split_connection_url_into_storage_account_and_container_name(connection_url)
        )
        self.blob_sas_url = f"https://{self.storage_account}.blob.core.windows.net?{self.sas_token}"
        self.client = BlobServiceClient(account_url=self.blob_sas_url)
        self.storage_bucket = self.client.get_container_client(self.container_name)

    @staticmethod
    def _split_connection_url_into_storage_account_and_container_name(
        connection_url: str,
    ) -> Tuple[str, str]:
        """Split the connection url into storage account and container name."""
        split_value = ".blob.core.windows.net"
        url_connection = urlparse(connection_url)
        storage_account = url_connection.hostname.split(split_value)[0]  # type: ignore
        container_name = url_connection.path.lstrip("/")
        return storage_account, container_name

    def get_blob_paths(self) -> List[str]:
        """List files in the Azure bucket."""
        return list(self.storage_bucket.list_blob_names())

    def get_blob_paths_as_tree(self) -> Dict:
        """Get a tree representation of the Azure bucket.

        Folder structure is represented as a dictionary.
        A folder is represented as a key with a value as a dictionary.
        A file is represented as a key with a value as None.
        """
        filetree = {}

        for filepath in self.storage_bucket.list_blob_names():
            current_node = filetree
            *folders, filename = Path(filepath).parts
            for folder in folders:
                if folder not in current_node:
                    current_node[folder] = {}
                current_node = current_node[folder]
            if filename:
                current_node[filename] = None

        return filetree


def get_blob_paths_azure_data_connection_with_service_credentials(
    data_integration: Dict, data_connection: Dict
) -> List[str]:
    """Get the blob paths for an Azure data connection using service credentials."""
    azure_client = AzureBucket(
        sas_token=data_integration["azureSASToken"],
        connection_url=data_integration["azureConnectionURL"],
    )

    blob_paths = azure_client.get_blob_paths()

    # blob_paths_in_bucket contains all blob paths in the bucket, we need to filter them
    # to keep only the ones in the data connection selected folders
    if isinstance(data_connection["selectedFolders"], List):
        blob_paths = [
            blob_path
            for blob_path in blob_paths
            if any(
                blob_path.startswith(selected_folder)
                for selected_folder in data_connection["selectedFolders"]
            )
        ]

    return blob_paths



"""Constants for the asset_import services."""

mime_extensions = {
    "Audio": "audio/x-flac,audio/mpeg,video/mp4",
    "Csv": "text/csv",
    "Video": (
        "video/mp4,video/x-matroska,video/3gpp,video/x-msvideo,"
        "video/x-m4v,video/quicktime,video/webm"
    ),
    "Image": (
        "image/jpeg,image/png,image/bmp,image/gif,image/webp,image/x-icon,"
        "image/tiff,image/vnd.microsoft.icon,image/svg+xml,image/avif,image/apng"
    ),
    "Pdf": "application/pdf",
    "Text": "text/plain",
    "TimeSeries": "text/csv",
}

project_compatible_mimetypes = {
    "AUDIO": mime_extensions["Audio"],
    "IMAGE": mime_extensions["Image"],
    "NA": "",
    "PDF": mime_extensions["Pdf"],
    "TEXT": mime_extensions["Text"],
    "TIME_SERIES": mime_extensions["Csv"],
    "URL": "",
    "VIDEO": mime_extensions["Video"],
    "VIDEO_LEGACY": "",
}

IMPORT_BATCH_SIZE = 100
FRAME_IMPORT_BATCH_SIZE = 1

MB_SIZE = 1024**2
LARGE_IMAGE_THRESHOLD_SIZE = 30 * MB_SIZE



"""Common code for the Kili exporter."""

import json
import os
from pathlib import Path
from typing import List

from kili.orm import Asset
from kili.services.export.format.base import AbstractExporter

from ...media.video import cut_video


class KiliExporter(AbstractExporter):
    """Common code for Kili exporters."""

    ASSETS_DIR_NAME = "assets"

    def _check_arguments_compatibility(self):
        """Checks if the export label format is compatible with the export options."""

    def _check_project_compatibility(self) -> None:
        """Checks if the export label format is compatible with the project."""

    def _save_assets_export(self, assets: List[Asset], output_filename: Path):
        """Save the assets to a file and return the link to that file."""
        self.logger.info("Exporting to kili format...")

        if self.with_assets:
            if self.project_input_type == "VIDEO":
                assets = self._cut_video_assets(assets)

            assets = self._clean_filepaths(assets)

        if self.single_file:
            project_json = json.dumps(assets, sort_keys=True, indent=4)
            self.base_folder.mkdir(parents=True, exist_ok=True)
            with (self.base_folder / "data.json").open("wb") as output_file:
                output_file.write(project_json.encode("utf-8"))
        else:
            labels_folder = self.base_folder / "labels"
            labels_folder.mkdir(parents=True, exist_ok=True)
            for asset in assets:
                external_id = asset["externalId"].replace(" ", "_")
                asset_json = json.dumps(asset, sort_keys=True, indent=4)
                with (labels_folder / f"{external_id}.json").open("wb") as output_file:
                    output_file.write(asset_json.encode("utf-8"))

        self.create_readme_kili_file(self.export_root_folder)

        self.make_archive(self.export_root_folder, output_filename)

        self.logger.warning(output_filename)

    def _clean_filepaths(self, assets: List[Asset]):
        # pylint: disable=line-too-long
        """Remove TemporaryDirectory() prefix from filepaths in "jsonContent" and "content" fields."""
        for asset in assets:
            if os.path.isfile(asset["content"]):
                asset["content"] = str(Path(self.ASSETS_DIR_NAME) / Path(asset["content"]).name)

            json_content_list = []
            if isinstance(asset["jsonContent"], list):
                for filepath in asset["jsonContent"]:
                    if os.path.isfile(filepath):
                        json_content_list.append(
                            str(Path(self.ASSETS_DIR_NAME) / Path(filepath).name)
                        )
                asset["jsonContent"] = json_content_list
        return assets

    def _cut_video_assets(self, assets: List[Asset]):
        """Cut video assets into frames."""
        for asset in assets:
            if asset["jsonContent"] == "" and os.path.isfile(asset["content"]):
                nbr_frames = len(asset.get("latestLabel", {}).get("jsonResponse", {}))
                if nbr_frames == 0:
                    continue
                leading_zeros = len(str(nbr_frames))
                asset["jsonContent"] = cut_video(
                    video_path=asset["content"],
                    asset=asset,
                    leading_zeros=leading_zeros,
                    output_dir=self.images_folder,
                )
        return assets

    def process_and_save(self, assets: List[Asset], output_filename: Path):
        """Extract formatted annotations from labels and save the json in the buckets."""
        clean_assets = self.pre_process_assets(assets, self.label_format)
        return self._save_assets_export(
            clean_assets,
            output_filename,
        )

    @property
    def images_folder(self) -> Path:
        """Export images folder."""
        return self.base_folder / self.ASSETS_DIR_NAME



"""Module for the "boundingPoly" key parsing of an object detection job response."""

from typing import Dict, Iterator, List, Optional, Union

from typeguard import typechecked
from typing_extensions import Literal

from .exceptions import InvalidMutationError
from .types import NormalizedVertex, Project


class BoundingPoly:
    """Class for parsing an element of a boundingPoly list."""

    def __init__(
        self,
        bounding_poly_json: Dict[
            Literal["normalizedVertices"],
            Union[List[NormalizedVertex], List[List[NormalizedVertex]]],
        ],
        type_of_tool: Optional[Literal["rectangle", "polygon", "semantic"]],
    ) -> None:
        """Class for BoundingPoly parsing."""
        self._json_data: Dict[
            Literal["normalizedVertices"],
            Union[List[NormalizedVertex], List[List[NormalizedVertex]]],
        ] = {}
        self._type_of_tool = type_of_tool

        self.normalized_vertices = bounding_poly_json["normalizedVertices"]

    def __str__(self) -> str:
        """Returns the string representation of the boundingPoly object."""
        return str(self._json_data)

    def __repr__(self) -> str:
        """Returns the string representation of the boundingPoly object."""
        return repr(self._json_data)

    def as_dict(self) -> Dict:
        """Returns the boundingPoly object as a dict."""
        return self._json_data

    @property
    def normalized_vertices(
        self,
    ) -> Union[List[NormalizedVertex], List[List[NormalizedVertex]]]:
        """Returns the normalized vertices of the bounding polygon."""
        return self._json_data["normalizedVertices"]

    @normalized_vertices.setter
    @typechecked
    def normalized_vertices(
        self,
        normalized_vertices: Union[List[NormalizedVertex], List[List[NormalizedVertex]]],
    ) -> None:
        """Sets the normalized vertices of the bounding polygon.

        Args:
            normalized_vertices: List of normalized vertices for object detection tasks.
                Or a list of a list of normalized vertices for NER in PDF task.
        """
        # for object detection tasks type of tool is defined
        if self._type_of_tool:
            nb_vertices = len(normalized_vertices)

            if self._type_of_tool == "rectangle" and nb_vertices != 4:
                raise InvalidMutationError(
                    f"Bounding polygon with {nb_vertices} vertices is not a rectangle."
                )

            if self._type_of_tool == "polygon" and nb_vertices < 3:
                raise InvalidMutationError(
                    f"Bounding polygon with {nb_vertices} vertices is not a polygon."
                )

        vertices = normalized_vertices if self._type_of_tool else normalized_vertices[0]
        for vertex in vertices:
            assert isinstance(vertex, dict), f"Vertex {vertex} is not a dict."
            if not 0 <= vertex["x"] <= 1:
                raise ValueError(f"Vertex x coordinate with value {vertex['x']} is not in [0, 1].")
            if not 0 <= vertex["y"] <= 1:
                raise ValueError(f"Vertex y coordinate with value {vertex['y']} is not in [0, 1].")

        self._json_data["normalizedVertices"] = normalized_vertices


class BoundingPolyList:
    """Class for the boundingPoly list parsing."""

    def __init__(
        self,
        bounding_poly_list: List[
            Dict[
                Literal["normalizedVertices"],
                Union[List[NormalizedVertex], List[List[NormalizedVertex]]],
            ]
        ],
        project_info: Project,
        job_name: str,
        type_of_tool: Optional[Literal["rectangle", "polygon", "semantic"]],
    ) -> None:
        """Class for the boundingPoly list parsing.

        Args:
            bounding_poly_list: List of dicts representing bounding polygons.
            project_info: Information about the project.
            job_name: Name of the job.
            type_of_tool: Type of tool used to create the bounding poly instances.
        """
        self._bounding_poly_list: List[BoundingPoly] = []
        self._project_info = project_info
        self._job_name = job_name
        self._type_of_tool: Optional[Literal["rectangle", "polygon", "semantic"]] = type_of_tool

        self._job_interface = project_info["jsonInterface"][job_name]  # type: ignore

        for bounding_poly_dict in bounding_poly_list:
            self.add_bounding_poly(bounding_poly_dict)

    def _check_can_append_bounding_poly(self, bounding_poly: BoundingPoly) -> None:
        if "normalizedVertices" not in bounding_poly._json_data:  # pylint: disable=protected-access
            raise KeyError("Bounding polygon must have a 'normalizedVertices' key.")

        if self._job_interface["mlTask"] not in ("OBJECT_DETECTION", "NAMED_ENTITIES_RECOGNITION"):
            raise InvalidMutationError(
                "BoundingPolyList can only be mutated for an object detection or NER in PDF jobs."
            )

    @typechecked
    def add_bounding_poly(
        self,
        bounding_poly: Dict[
            Literal["normalizedVertices"],
            Union[List[NormalizedVertex], List[List[NormalizedVertex]]],
        ],
    ) -> None:
        """Adds a boundingPoly object to a BoundingPolyList object."""
        bounding_poly_obj = BoundingPoly(bounding_poly, type_of_tool=self._type_of_tool)
        self._check_can_append_bounding_poly(bounding_poly_obj)
        self._bounding_poly_list.append(bounding_poly_obj)

    @typechecked
    def __getitem__(self, index: int) -> BoundingPoly:
        """Returns the boundingPoly object at the given index."""
        return self._bounding_poly_list[index]

    def __len__(self) -> int:
        """Returns the number of bounding polygons."""
        return len(self._bounding_poly_list)

    def __iter__(self) -> Iterator[BoundingPoly]:
        """Allows iterating over the boundingPoly list."""
        return iter(self._bounding_poly_list)

    def __str__(self) -> str:
        """Returns the string representation of the boundingPoly list."""
        return str(self.as_list())

    def __repr__(self) -> str:
        """Returns the string representation of the boundingPoly list."""
        return repr(self.as_list())

    def as_list(self) -> List[Dict]:
        """Returns the list of bounding polygons as a list of dicts."""
        return [bounding_poly.as_dict() for bounding_poly in self._bounding_poly_list]



"""Api key queries."""

from typing import Dict, Generator, Iterable, List, Optional, overload

from typeguard import typechecked
from typing_extensions import Literal

from kili.core.graphql import QueryOptions
from kili.core.graphql.graphql_client import GraphQLClient
from kili.core.graphql.operations.api_key.queries import APIKeyQuery, APIKeyWhere
from kili.core.helpers import disable_tqdm_if_as_generator
from kili.utils.logcontext import for_all_methods, log_call


@for_all_methods(log_call, exclude=["__init__"])
class QueriesApiKey:
    """Set of ApiKey queries."""

    graphql_client: GraphQLClient

    # pylint: disable=too-many-arguments,dangerous-default-value
    @overload
    def api_keys(
        self,
        api_key_id: Optional[str] = None,
        user_id: Optional[str] = None,
        api_key: Optional[str] = None,
        skip: int = 0,
        fields: List[str] = ["id", "name", "createdAt", "revoked"],
        first: Optional[int] = None,
        disable_tqdm: bool = False,
        *,
        as_generator: Literal[True],
    ) -> Generator[Dict, None, None]:
        ...

    @overload
    def api_keys(
        self,
        api_key_id: Optional[str] = None,
        user_id: Optional[str] = None,
        api_key: Optional[str] = None,
        skip: int = 0,
        fields: List[str] = ["id", "name", "createdAt", "revoked"],
        first: Optional[int] = None,
        disable_tqdm: bool = False,
        *,
        as_generator: Literal[False] = False,
    ) -> List[Dict]:
        ...

    @typechecked
    def api_keys(
        self,
        api_key_id: Optional[str] = None,
        user_id: Optional[str] = None,
        api_key: Optional[str] = None,
        skip: int = 0,
        fields: List[str] = ["id", "name", "createdAt", "revoked"],
        first: Optional[int] = None,
        disable_tqdm: bool = False,
        *,
        as_generator: bool = False,
    ) -> Iterable[Dict]:
        # pylint: disable=line-too-long
        """Get a generator or a list of API keys that match a set of constraints.

        !!! info
            You can only query your own API keys

        Args:
            api_key_id: Identifier of the API key to retrieve.
            user_id: Identifier of the user.
            api_key: Value of the API key.
            skip: Number of assets to skip (they are ordered by their date of creation, first to last).
            fields: All the fields to request among the possible fields for the assets.
                See [the documentation](https://docs.kili-technology.com/reference/graphql-api#apikey) for all possible fields.
            first: Maximum number of API keys to return.
            disable_tqdm: If `True`, the progress bar will be disabled.
            as_generator: If `True`, a generator on the API key is returned.

        Returns:
            A result object which contains the query if it was successful,
                or an error message.


        Examples:
            >>> kili.api_keys(user_id=user_id)
            >>> kili.api_keys(api_key=api_key)
            >>> kili.api_keys(api_key=api_key, as_generator=False)
        """
        where = APIKeyWhere(api_key_id=api_key_id, user_id=user_id, api_key=api_key)
        disable_tqdm = disable_tqdm_if_as_generator(as_generator, disable_tqdm)
        options = QueryOptions(disable_tqdm, first, skip)
        api_keys_gen = APIKeyQuery(self.graphql_client)(where, fields, options)

        if as_generator:
            return api_keys_gen
        return list(api_keys_gen)

    @typechecked
    def count_api_keys(
        self,
        api_key_id: Optional[str] = None,
        user_id: Optional[str] = None,
        api_key: Optional[str] = None,
    ) -> int:
        """Count and return the number of api keys with the given constraints.

        Args:
            api_key_id: Identifier of the API key to retrieve.
            user_id: Identifier of the user.
            api_key: Value of the api key.

        Returns:
            The number of API Keys matching params if it was successful,
                or an error message.

        Examples:
            >>> kili.count_api_keys(user_id=user_id)
            3
            >>> kili.count_api_keys(api_key=api_key)
            1
        """
        where = APIKeyWhere(api_key_id=api_key_id, user_id=user_id, api_key=api_key)
        return APIKeyQuery(self.graphql_client).count(where)



"""CLI's project export subcommand."""

from typing import Optional, cast

import click
from typeguard import typechecked
from typing_extensions import get_args

from kili import services
from kili.entrypoints.cli.common_args import Options
from kili.entrypoints.cli.helpers import get_kili_client
from kili.services.export.exceptions import NoCompatibleJobError
from kili.services.export.types import LabelFormat, SplitOption
from kili.services.types import ProjectId


@click.command(name="export")
@click.option(
    "--output-format",
    type=click.Choice(get_args(LabelFormat)),
    help="Format into which the label data will be converted",
    required=True,
)
@click.option(
    "--output-file", type=str, help="File into which the labels are saved.", required=True
)
@click.option(
    "--layout",
    type=click.Choice(get_args(SplitOption)),
    default="merged",
    help=(
        "Layout of the label files: 'split' to group labels per job, 'merged' to have one folder"
        " with every labels."
    ),
)
@click.option(
    "--single-file",
    type=bool,
    is_flag=True,
    help=(
        "Layout of the label files. Single file mode is only available for some specific formats"
        " (COCO and Kili)."
    ),
)
@click.option(
    "--with-assets/--without-assets",
    type=bool,
    default=True,
    help="Download assets in the export.",
)
@Options.api_key
@Options.endpoint
@Options.project_id
@Options.verbose
@typechecked
# pylint: disable=too-many-arguments
def export_labels(
    output_format: LabelFormat,
    output_file: str,
    layout: SplitOption,
    single_file: bool,
    api_key: Optional[str],
    endpoint: Optional[str],
    project_id: str,
    verbose: bool,
    with_assets: bool,
):
    # pylint: disable=line-too-long
    """Export the Kili labels of a project to a given format.

    \b
    !!! Info
        The supported formats are:

        - Yolo V4, V5, V7 for object detection tasks (bounding box).
        - Kili (a.k.a raw) for all tasks.
        - COCO for object detection tasks (bounding box and semantic segmentation).
        - Pascal VOC for object detection tasks (bounding box).
    \b
    \b
    !!! warning "Cloud storage"
        Export with asset download (`--with-assets`) is not allowed for projects connected to a cloud storage.
    \b
    \b
    !!! Examples
        ```
        kili project export \\
            --project-id <project_id> \\
            --output-format coco \\
            --output-file /tmp/export.zip
        ```
        ```
        kili project export \\
            --project-id <project_id> \\
            --output-format yolo_v5 \\
            --output-file /tmp/export_split.zip \\
            --layout split
        ```
    """
    kili = get_kili_client(api_key=api_key, api_endpoint=endpoint)

    try:
        services.export_labels(
            kili,
            asset_ids=None,
            project_id=cast(ProjectId, project_id),
            export_type="latest",
            label_format=output_format,
            split_option=layout,
            single_file=single_file,
            output_file=output_file,
            disable_tqdm=not verbose,
            log_level="INFO" if verbose else "WARNING",
            with_assets=with_assets,
            annotation_modifier=None,
            asset_filter_kwargs=None,
        )
    except NoCompatibleJobError as excp:
        print(str(excp))



"""GraphQL Queries of Users."""


from typing import Optional

from kili.core.graphql import BaseQueryWhere, GraphQLQuery


class UserWhere(BaseQueryWhere):
    """Tuple to be passed to the UserQuery to restrict query."""

    def __init__(
        self,
        api_key: Optional[str] = None,
        email: Optional[str] = None,
        organization_id: Optional[str] = None,
    ):
        self.api_key = api_key
        self.email = email
        self.organization_id = organization_id
        super().__init__()

    def graphql_where_builder(self):
        """Build the GraphQL Where payload sent in the resolver from the SDK UserWhere."""
        return {
            "apiKey": self.api_key,
            "email": self.email,
            "organization": {"id": self.organization_id},
        }


class UserQuery(GraphQLQuery):
    """User query."""

    @staticmethod
    def query(fragment):
        """Return the GraphQL users query."""
        return f"""
        query users($where: UserWhere!, $first: PageSize!, $skip: Int!) {{
            data: users(where: $where, first: $first, skip: $skip) {{
                {fragment}
            }}
        }}
        """

    COUNT_QUERY = """
    query countUsers($where: UserWhere!) {
        data: countUsers(where: $where)
    }
    """


GQL_ME = """
query Me {
    data: me {
        id
        email
    }
}
"""



"""Services for data connections."""
import logging
import time
from typing import Any, Dict, List

from tenacity import Retrying
from tenacity.retry import retry_if_exception_type
from tenacity.stop import stop_after_delay
from tenacity.wait import wait_exponential
from typing_extensions import Literal

from kili.core.graphql import QueryOptions
from kili.core.graphql.operations.asset.queries import AssetQuery, AssetWhere
from kili.core.graphql.operations.data_connection.queries import (
    DataConnectionIdWhere,
    DataConnectionQuery,
)
from kili.core.helpers import format_result
from kili.entrypoints.mutations.data_connection.queries import (
    GQL_COMPUTE_DATA_CONNECTION_DIFFERENCES,
    GQL_VALIDATE_DATA_DIFFERENCES,
)

LOGGER = None


def _get_logger() -> logging.Logger:
    global LOGGER  # pylint: disable=global-statement

    if LOGGER is not None:
        return LOGGER

    LOGGER = logging.getLogger(__name__)
    LOGGER.setLevel(logging.INFO)
    LOGGER.addHandler(logging.StreamHandler())
    return LOGGER


def get_data_connection(kili, data_connection_id: str, fields: List[str]) -> Dict:
    """Get data connection information."""
    where = DataConnectionIdWhere(data_connection_id=data_connection_id)
    options = QueryOptions(first=1, disable_tqdm=True)
    data_connection = list(DataConnectionQuery(kili.graphql_client)(where, fields, options))
    if len(data_connection) == 0:
        raise ValueError(f"No data connection with id {data_connection_id}")
    return data_connection[0]


def trigger_validate_data_differences(
    kili, diff_type: Literal["ADD", "REMOVE"], data_connection_id: str
) -> Dict:
    """Call the validateDataDifferences resolver."""
    variables = {
        "where": {"connectionId": data_connection_id, "type": diff_type},
        "processingParameters": None,
    }
    result = kili.graphql_client.execute(GQL_VALIDATE_DATA_DIFFERENCES, variables)
    data_connection = format_result("data", result)
    return data_connection


def validate_data_differences(
    kili, diff_type: Literal["ADD", "REMOVE"], data_connection: Dict
) -> None:
    """Call the validateDataDifferences resolver and wait until the validation is done."""
    diff = data_connection["dataDifferencesSummary"]["added" if diff_type == "ADD" else "removed"]

    where = AssetWhere(project_id=data_connection["projectId"])
    nb_assets_before = AssetQuery(kili.graphql_client).count(where)

    trigger_validate_data_differences(kili, diff_type, data_connection["id"])

    for attempt in Retrying(
        wait=wait_exponential(multiplier=1, min=1, max=4),
        stop=stop_after_delay(5 * 60),
        retry=retry_if_exception_type(ValueError),
        reraise=True,
    ):
        with attempt:
            nb_assets_after = AssetQuery(kili.graphql_client).count(where)
            if abs(nb_assets_after - nb_assets_before) != diff:
                raise ValueError(
                    "Number of assets in project after validation is not correct: before"
                    f" {nb_assets_before} assets, after {nb_assets_after} assets,"
                    f" dataDifferencesSummary diff {diff}"
                )


def compute_differences(kili, data_connection_id: str) -> Dict:
    """Compute the data connection differences."""
    logger = _get_logger()
    logger.info("Computing the data connection differences.")

    data_connection = get_data_connection(
        kili,
        data_connection_id,
        fields=[
            "dataIntegration.azureIsUsingServiceCredentials",
            "dataIntegration.platform",
            "dataIntegration.azureSASToken",
            "dataIntegration.azureConnectionURL",
            "dataIntegration.id",
            "selectedFolders",
        ],
    )

    data_integration = data_connection["dataIntegration"]

    blob_paths = None

    # for azure using credentials, it is required to provide the blob paths to compute the diffs
    if (
        data_integration["platform"] == "Azure"
        and data_integration["azureIsUsingServiceCredentials"]
    ):
        logger.info("Azure data integration is using service credentials. Retrieving blob paths.")
        if not (data_integration["azureSASToken"] and data_integration["azureConnectionURL"]):
            raise ValueError(
                f"Cannot compute differences for data connection {data_connection_id} with data"
                f" integration {data_integration['id']}. Need to provide \"azureSASToken\" and"
                f' "azureConnectionURL" in data integration: {data_integration}'
            )

        try:
            # pylint: disable=import-outside-toplevel
            from .azure import (
                get_blob_paths_azure_data_connection_with_service_credentials,
            )
        except ImportError as err:
            raise ImportError(
                "The azure-storage-blob package is required to use Azure buckets. "
                " Run `pip install kili[azure]` to install it."
            ) from err

        blob_paths = get_blob_paths_azure_data_connection_with_service_credentials(
            data_connection=data_connection, data_integration=data_integration
        )

    variables: Dict[str, Any] = {"where": {"id": data_connection_id}}
    if blob_paths is not None:
        variables["data"] = {"blobPaths": blob_paths}
    result = kili.graphql_client.execute(GQL_COMPUTE_DATA_CONNECTION_DIFFERENCES, variables)
    data_connection = format_result("data", result)
    return data_connection


def verify_diff_computed(kili, data_connection_id: str) -> None:
    """Verify that the data connection differences have been computed.

    Trigger the computation if not already computing.
    """
    logger = _get_logger()
    logger.info("Verifying that the data connection differences have been computed.")

    data_connection = get_data_connection(kili, data_connection_id, fields=["isChecking"])

    if not data_connection["isChecking"]:
        compute_differences(kili, data_connection_id)
        time.sleep(1)  # backend needs some time to refresh the "isChecking"

    for attempt in Retrying(
        wait=wait_exponential(multiplier=1, min=1, max=4),
        stop=stop_after_delay(5 * 60),
        retry=retry_if_exception_type(ValueError),
        reraise=True,
    ):
        with attempt:
            data_connection = get_data_connection(kili, data_connection_id, fields=["isChecking"])
            if data_connection["isChecking"]:
                raise ValueError(f"Data connection is still checking: {data_connection}")

    time.sleep(1)  # backend needs some time to refresh the "isChecking"


def synchronize_data_connection(
    kili, data_connection_id: str, delete_extraneous_files: bool, dry_run: bool
) -> Dict:
    """Launch a data connection synchronization."""
    logger = _get_logger()
    logger.info("Synchronizing data connection: %s", data_connection_id)

    verify_diff_computed(kili, data_connection_id)

    data_connection = get_data_connection(
        kili,
        data_connection_id,
        fields=[
            "id",
            "dataDifferencesSummary.added",
            "dataDifferencesSummary.removed",
            "dataDifferencesSummary.total",
            "isChecking",
            "numberOfAssets",
            "projectId",
        ],
    )
    if data_connection["isChecking"]:
        raise ValueError(f"Data connection should not be checking: {data_connection}")

    added = data_connection["dataDifferencesSummary"]["added"]
    removed = data_connection["dataDifferencesSummary"]["removed"]
    total = data_connection["dataDifferencesSummary"]["total"]
    nb_assets = data_connection["numberOfAssets"]

    logger.info("Currently %d asset(s) imported from the data connection.", nb_assets)

    if total == 0:
        logger.info("No differences found. Nothing to synchronize.")
        return data_connection

    logger.info(
        "Found %d difference(s): %d asset(s) to add, %d asset(s) to remove.", total, added, removed
    )

    if dry_run:
        # pylint: disable=unnecessary-lambda-assignment
        validate_data_differences_func = lambda *args, **kwargs: None  # noqa: E731
        logger.info("Dry run: no data will be added or removed.")
    else:
        validate_data_differences_func = validate_data_differences

    if removed > 0:
        if delete_extraneous_files:
            validate_data_differences_func(kili, "REMOVE", data_connection)
            logger.info("Removed %d extraneous file(s).", removed)
        else:
            logger.info(
                "Use delete_extraneous_files=True to remove %d extraneous file(s).",
                removed,
            )

    if added > 0:
        validate_data_differences_func(kili, "ADD", data_connection)
        logger.info("Added %d file(s).", added)

    return get_data_connection(kili, data_connection_id, fields=["numberOfAssets", "projectId"])
